# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from pyxem.generators.calibration_generator import CalibrationGenerator

from pyxem.signals.electron_diffraction2d import ElectronDiffraction2D

from pyxem.utils.calibration_utils import (
    call_ring_pattern,
    calc_radius_with_distortion,
    generate_ring_pattern,
)


@pytest.fixture
def input_parameters():
    x0 = [95, 1200, 2.8, 450, 1.5, 10]
    return x0


@pytest.mark.parametrize(
    "known_values",
    [
        (
            np.array(
                [
                    124.05909278,
                    25.85258647,
                    39.09906246,
                    173.75469207,
                    79.48046629,
                    533.72925614,
                    36.23521052,
                    29.58603406,
                    21.83270633,
                    75.89239623,
                    40.04732689,
                    14.52041808,
                    35.82637996,
                    75.33666451,
                    21.21751965,
                    38.97731538,
                    19.64631964,
                    161.72783637,
                    23.6894442,
                    282.3126376,
                ]
            )
        )
    ],
)
@pytest.mark.parametrize(
    "reference_indices",
    [
        (
            np.array(
                [
                    [205, 158],
                    [197, 1],
                    [105, 239],
                    [64, 148],
                    [61, 84],
                    [136, 155],
                    [37, 85],
                    [21, 94],
                    [247, 31],
                    [171, 195],
                    [202, 39],
                    [225, 255],
                    [233, 128],
                    [56, 107],
                    [22, 51],
                    [28, 119],
                    [20, 45],
                    [164, 65],
                    [235, 188],
                    [75, 186],
                ]
            )
        )
    ],
)
def test_generate_ring_pattern(input_parameters, known_values, reference_indices):
    x0 = input_parameters
    rings = generate_ring_pattern(
        image_size=256,
        mask=True,
        mask_radius=10,
        scale=x0[0],
        amplitude=x0[1],
        spread=x0[2],
        direct_beam_amplitude=x0[3],
        asymmetry=x0[4],
        rotation=x0[5],
    )
    assert np.allclose(
        known_values, rings[reference_indices[:, 0], reference_indices[:, 1]]
    )



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
from pyxem.signals.diffraction_vectors import DiffractionVectors
from sklearn.cluster import DBSCAN
from hyperspy.signals import Signal2D

# DiffractionVectors correspond to a single list of vectors, a map of vectors
# all of equal length, and the ragged case. A fixture is defined for each of
# these cases and all methods tested for it.


@pytest.fixture(
    params=[
        np.array(
            [
                [0.063776, 0.011958],
                [-0.035874, 0.131538],
                [0.035874, -0.131538],
                [0.035874, 0.143496],
                [-0.035874, -0.13951],
                [-0.115594, 0.123566],
                [0.103636, -0.11958],
                [0.123566, 0.151468],
            ]
        )
    ]
)
def diffraction_vectors_single(request):
    dvs = DiffractionVectors(request.param)
    dvs.axes_manager.set_signal_dimension(1)
    return dvs


@pytest.fixture(
    params=[
        np.array(
            [
                [
                    np.array(
                        [
                            [0.089685, 0.292971],
                            [0.017937, 0.277027],
                            [-0.069755, 0.257097],
                            [-0.165419, 0.241153],
                            [0.049825, 0.149475],
                            [-0.037867, 0.129545],
                            [-0.117587, 0.113601],
                        ]
                    ),
                    np.array(
                        [
                            [0.089685, 0.292971],
                            [0.017937, 0.277027],
                            [-0.069755, 0.257097],
                            [-0.165419, 0.241153],
                            [0.049825, 0.149475],
                            [-0.037867, 0.129545],
                            [-0.117587, 0.113601],
                            [0.149475, 0.065769],
                            [0.229195, 0.045839],
                            [0.141503, 0.025909],
                            [0.073741, 0.013951],
                            [0.001993, 0.001993],
                            [-0.069755, -0.009965],
                        ]
                    ),
                ],
                [
                    np.array(
                        [
                            [0.089685, 0.292971],
                            [0.017937, 0.277027],
                            [-0.069755, 0.257097],
                            [-0.165419, 0.241153],
                            [0.049825, 0.149475],
                            [-0.037867, 0.129545],
                            [-0.117587, 0.113601],
                            [0.149475, 0.065769],
                            [0.229195, 0.045839],
                            [0.141503, 0.025909],
                            [0.073741, 0.013951],
                        ]
                    ),
                    np.array([[0.001993, 0.001993]]),
                ],
            ],
            dtype=object,
        )
    ]
)
def diffraction_vectors_map(request):
    dvm = DiffractionVectors(request.param)
    dvm.axes_manager.set_signal_dimension(0)
    dvm.axes_manager[0].name = "x"
    dvm.axes_manager[1].name = "y"
    return dvm


def test_plot_diffraction_vectors(diffraction_vectors_map):
    with pytest.warns(UserWarning, match="distance_threshold=0 was given"):
        diffraction_vectors_map.plot_diffraction_vectors(
            xlim=1.0, ylim=1.0, distance_threshold=0
        )


def test_plot_diffraction_vectors_on_signal(
    diffraction_vectors_map, diffraction_pattern
):
    diffraction_vectors_map.plot_diffraction_vectors_on_signal(diffraction_pattern)


def test_get_cartesian_coordinates(diffraction_vectors_map):
    accelerating_voltage = 200
    camera_length = 0.2
    diffraction_vectors_map.calculate_cartesian_coordinates(
        accelerating_voltage, camera_length
    )
    # Coordinate conversion is tested in vector_utils. Just test that the
    # result is stored correctly
    assert diffraction_vectors_map.cartesian is not None
    assert (
        diffraction_vectors_map.axes_manager[0].name
        == diffraction_vectors_map.cartesian.axes_manager[0].name
    )


class TestMagnitudes:
    def test_get_magnitudes_single(self, diffraction_vectors_single):
        diffraction_vectors_single.get_magnitudes()

    @pytest.mark.filterwarnings("ignore::FutureWarning")  # deemed "safe enough"
    def test_get_magnitude_histogram_single(self, diffraction_vectors_single):
        diffraction_vectors_single.get_magnitude_histogram(bins=np.arange(0, 0.5, 0.1))

    def test_get_magnitudes_map(self, diffraction_vectors_map):
        diffraction_vectors_map.get_magnitudes()

    @pytest.mark.filterwarnings("ignore::FutureWarning")  # deemed "safe enough"
    def test_get_magnitude_histogram_map(self, diffraction_vectors_map):
        diffraction_vectors_map.get_magnitude_histogram(bins=np.arange(0, 0.5, 0.1))


class TestUniqueVectors:
    def test_get_unique_vectors_map_type(self, diffraction_vectors_map):
        unique_vectors = diffraction_vectors_map.get_unique_vectors()
        assert isinstance(unique_vectors, DiffractionVectors)

    def test_get_unique_vectors_single(self, diffraction_vectors_single):
        diffraction_vectors_single.get_unique_vectors()

    @pytest.mark.parametrize(
        "distance_threshold, answer",
        [
            (
                0.01,
                np.array(
                    [
                        [-0.165419, 0.241153],
                        [-0.117587, 0.113601],
                        [-0.069755, -0.009965],
                        [-0.069755, 0.257097],
                        [-0.037867, 0.129545],
                        [0.001993, 0.001993],
                        [0.017937, 0.277027],
                        [0.049825, 0.149475],
                        [0.073741, 0.013951],
                        [0.089685, 0.292971],
                        [0.141503, 0.025909],
                        [0.149475, 0.065769],
                        [0.229195, 0.045839],
                    ]
                ),
            ),
            (
                0.1,
                np.array(
                    [
                        [-0.117587, 0.249125],
                        [-0.077727, 0.121573],
                        [-0.021923, -0.001993],
                        [0.053811, 0.284999],
                        [0.049825, 0.149475],
                        [0.121573, 0.03520967],
                        [0.229195, 0.045839],
                    ]
                ),
            ),
        ],
    )
    def test_get_unique_vectors_map_values(
        self, diffraction_vectors_map, distance_threshold, answer
    ):
        unique_vectors = diffraction_vectors_map.get_unique_vectors(
            distance_threshold=distance_threshold
        )
        np.testing.assert_almost_equal(unique_vectors.data, answer)

    def test_get_unique_vectors_map_dbscan(self, diffraction_vectors_map):
        unique_dbscan = diffraction_vectors_map.get_unique_vectors(
            method="DBSCAN", return_clusters=True
        )
        assert isinstance(unique_dbscan[0], DiffractionVectors)
        assert isinstance(unique_dbscan[1], DBSCAN)

    @pytest.mark.parametrize(
        "distance_threshold, answer",
        [
            (
                0.01,
                np.array(
                    [
                        [-0.165419, 0.241153],
                        [-0.117587, 0.113601],
                        [-0.069755, -0.009965],
                        [-0.069755, 0.257097],
                        [-0.037867, 0.129545],
                        [0.001993, 0.001993],
                        [0.017937, 0.277027],
                        [0.049825, 0.149475],
                        [0.073741, 0.013951],
                        [0.089685, 0.292971],
                        [0.141503, 0.025909],
                        [0.149475, 0.065769],
                        [0.229195, 0.045839],
                    ]
                ),
            ),
            (
                0.1,
                np.array(
                    [
                        [-0.031888, 0.267062],
                        [-0.03520967, 0.13087367],
                        [0.10200536, 0.02699609],
                    ]
                ),
            ),
        ],
    )
    def test_get_unique_vectors_map_values_dbscan(
        self, diffraction_vectors_map, distance_threshold, answer
    ):
        unique_vectors = diffraction_vectors_map.get_unique_vectors(
            distance_threshold=distance_threshold, method="DBSCAN"
        )
        np.testing.assert_almost_equal(unique_vectors.data, answer)


class TestFilterVectors:
    def test_filter_magnitude_map_type(self, diffraction_vectors_map):
        filtered_vectors = diffraction_vectors_map.filter_magnitude(0.1, 1.0)
        assert isinstance(filtered_vectors, DiffractionVectors)

    def test_filter_magnitude_single_type(self, diffraction_vectors_single):
        filtered_vectors = diffraction_vectors_single.filter_magnitude(0.1, 1.0)
        assert isinstance(filtered_vectors, DiffractionVectors)

    def test_filter_magnitude_map(self, diffraction_vectors_map):
        filtered_vectors = diffraction_vectors_map.filter_magnitude(0.1, 1.0)
        ans = np.array(
            [
                [0.089685, 0.292971],
                [0.017937, 0.277027],
                [-0.069755, 0.257097],
                [-0.165419, 0.241153],
                [0.049825, 0.149475],
                [-0.037867, 0.129545],
                [-0.117587, 0.113601],
                [0.149475, 0.065769],
                [0.229195, 0.045839],
                [0.141503, 0.025909],
            ]
        )
        np.testing.assert_almost_equal(filtered_vectors.data[0][1], ans)

    def test_filter_magnitude_single(self, diffraction_vectors_single):
        filtered_vectors = diffraction_vectors_single.filter_magnitude(0.15, 1.0)
        ans = np.array(
            [[-0.115594, 0.123566], [0.103636, -0.11958], [0.123566, 0.151468]]
        )
        np.testing.assert_almost_equal(filtered_vectors.data, ans)

    def test_filter_detector_edge_map_type(self, diffraction_vectors_map):
        diffraction_vectors_map.detector_shape = (260, 240)
        diffraction_vectors_map.pixel_calibration = 0.001
        filtered_vectors = diffraction_vectors_map.filter_detector_edge(exclude_width=2)
        assert isinstance(filtered_vectors, DiffractionVectors)

    def test_filter_detector_edge_single_type(self, diffraction_vectors_single):
        diffraction_vectors_single.detector_shape = (260, 240)
        diffraction_vectors_single.pixel_calibration = 0.001
        filtered_vectors = diffraction_vectors_single.filter_detector_edge(
            exclude_width=10
        )
        assert isinstance(filtered_vectors, DiffractionVectors)

    def test_filter_detector_edge_map(self, diffraction_vectors_map):
        diffraction_vectors_map.detector_shape = (260, 240)
        diffraction_vectors_map.pixel_calibration = 0.001
        filtered_vectors = diffraction_vectors_map.filter_detector_edge(exclude_width=2)
        ans = np.array([[-0.117587, 0.113601]])
        np.testing.assert_almost_equal(filtered_vectors.data[0, 0], ans)

    def test_filter_detector_edge_single(self, diffraction_vectors_single):
        diffraction_vectors_single.detector_shape = (260, 240)
        diffraction_vectors_single.pixel_calibration = 0.001
        filtered_vectors = diffraction_vectors_single.filter_detector_edge(
            exclude_width=10
        )
        ans = np.array([[0.063776, 0.011958]])
        np.testing.assert_almost_equal(filtered_vectors.data, ans)


class TestDiffractingPixelsMap:
    def test_get_dpm_values(self, diffraction_vectors_map):
        answer = np.array([[7.0, 13.0], [11.0, 1.0]])
        xim = diffraction_vectors_map.get_diffracting_pixels_map()
        assert np.allclose(xim, answer)

    def test_get_dpm_type(self, diffraction_vectors_map):
        xim = diffraction_vectors_map.get_diffracting_pixels_map()
        assert isinstance(xim, Signal2D)

    def test_get_dpm_title(self, diffraction_vectors_map):
        xim = diffraction_vectors_map.get_diffracting_pixels_map()
        assert xim.metadata.General.title == "Diffracting Pixels Map"

    def test_get_dpm_in_range(self, diffraction_vectors_map):
        answer = np.array([[0.0, 3.0], [1.0, 1.0]])
        xim = diffraction_vectors_map.get_diffracting_pixels_map(in_range=(0, 0.1))
        assert np.allclose(xim, answer)

    def test_get_dpm_binary(self, diffraction_vectors_map):
        answer = np.array([[1.0, 1.0], [1.0, 1.0]])
        xim = diffraction_vectors_map.get_diffracting_pixels_map(binary=True)
        assert np.allclose(xim, answer)



import numpy as np


def _correlation(z, axis=0, mask=None, wrap=True, normalize=True):
    """A generic function for applying a correlation with a mask.

     Takes a nd image and then preforms a auto-correlation on some axis.
     Used in the electron correlation and angular correlation codes. Uses
     the fft to speed up the correlation.
    Parameters
    ----------
    z: np.array
        A nd numpy array
    axis: int
        The axis to apply the correlation to
    mask: np.array
        A boolean array of the same size as z
    wrap: bool
        Allow the function to wrap or add zeros to the beginning and the
        end of z on the specified axis
    normalize: bool
        Subtract <I(\theta)>^2 and divide by <I(\theta)>^2
    """
    if wrap is False:
        z_shape = np.shape(z)
        padder = [(0, 0)] * len(z_shape)
        pad = z_shape[axis]  #  This will be faster if the length of the axis
        # is a power of 2.  Based on the numpy implementation.  Not terribly
        # faster I think..
        padder[axis] = (pad, pad)
        slicer = [slice(None),] * len(z_shape)
        slicer[axis] = slice(0, -2 * pad)  # creating the proper slices
        if mask is None:
            mask = np.zeros(shape=np.shape(z))
        z = np.pad(z, padder, "constant")

    if mask is not None or wrap is False:  # Need to scale for wrapping properly
        m = np.array(mask, dtype=bool)  # make sure it is a boolean array
        # this is to determine how many of the variables were non zero... This is really dumb.  but...
        # it works and I should stop trying to fix it (wreak it)
        mask_boolean = ~m  # inverting the boolean mask
        if wrap is False:  # padding with zeros to the function along some axis
            m = np.pad(
                m, padder, "constant"
            )  # all the zeros are masked (should account for padding
            #  when normalized.
            mask_boolean = np.pad(mask_boolean, padder, "constant")
        mask_fft = np.fft.rfft(mask_boolean, axis=axis)
        number_unmasked = np.fft.irfft(
            mask_fft * np.conjugate(mask_fft), axis=axis
        ).real
        number_unmasked[
            number_unmasked < 1
        ] = 1  # get rid of divide by zero error for completely masked rows
        z[m] = 0

    # fast method uses a FFT and is a process which is O(n) = n log(n)
    I_fft = np.fft.rfft(z, axis=axis)
    a = np.fft.irfft(I_fft * np.conjugate(I_fft), axis=axis)

    if mask is not None:
        a = np.multiply(np.divide(a, number_unmasked), np.shape(z)[0])

    if normalize:  # simplified way to calculate the normalization
        row_mean = np.mean(a, axis=axis)
        row_mean[row_mean == 0] = 1
        row_mean = np.expand_dims(row_mean, axis=axis)
        a = np.divide(np.subtract(a, row_mean), row_mean)

    if wrap is False:
        print(slicer)
        a = a[slicer]
    return a


def _power(z, axis=0, mask=None, wrap=True, normalize=True):
    """The power spectrum of the correlation.

    This method is a little more complex if mask is not None due to
    the extra calculations necessary to ignore some of the pixels
    during the calculations.

    Parameters
    ----------------
    z: np.array
        Some n-d array to get the power spectrum from.
    axis: int
        The axis to preform the operation on.
    mask: np.array
        A boolean mask to be applied.
    wrap: bool
        Choose if the function should wrap.  In most cases this will be True
        when calculating the power of some function
    normalize: bool
        Choose to normalize the function by the mean.

    Returns
    -----------------
    power: np.array
        The power spectrum along some axis
    """
    if mask is None:  # This might not normalize things as well
        I_fft = np.fft.rfft(z, axis=axis)
        return (I_fft * np.conjugate(I_fft)).real
    else:
        return np.power(
            np.fft.rfft(
                _correlation(z=z, axis=axis, mask=mask, wrap=wrap, normalize=normalize)
            ),
            2,
        ).real


def corr_to_power(z):
    return np.fft.rfft(z, axis=1).real



# -*- coding: utf-8 -*-
# Copyright 2017-2019 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import dask.array as da

from hyperspy.signals import Signal2D

from pyxem.signals.polar_diffraction2d import PolarDiffraction2D, LazyPolarDiffraction2D
from pyxem.signals.correlation2d import Correlation2D
from pyxem.signals.power2d import Power2D


class TestComputeAndAsLazy2D:
    def test_2d_data_compute(self):
        dask_array = da.random.random((100, 150), chunks=(50, 50))
        s = LazyPolarDiffraction2D(dask_array)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s.compute()
        assert s.__class__ == PolarDiffraction2D
        assert not hasattr(s.data, "compute")
        assert s.axes_manager[0].scale == scale0
        assert s.axes_manager[1].scale == scale1
        assert s.metadata.Test == metadata_string
        assert dask_array.shape == s.data.shape

    def test_4d_data_compute(self):
        dask_array = da.random.random((4, 4, 10, 15), chunks=(1, 1, 10, 15))
        s = LazyPolarDiffraction2D(dask_array)
        s.compute()
        assert s.__class__ == PolarDiffraction2D
        assert dask_array.shape == s.data.shape

    def test_2d_data_as_lazy(self):
        data = np.random.random((100, 150))
        s = PolarDiffraction2D(data)
        scale0, scale1, metadata_string = 0.5, 1.5, "test"
        s.axes_manager[0].scale = scale0
        s.axes_manager[1].scale = scale1
        s.metadata.Test = metadata_string
        s_lazy = s.as_lazy()
        assert s_lazy.__class__ == LazyPolarDiffraction2D
        assert hasattr(s_lazy.data, "compute")
        assert s_lazy.axes_manager[0].scale == scale0
        assert s_lazy.axes_manager[1].scale == scale1
        assert s_lazy.metadata.Test == metadata_string
        assert data.shape == s_lazy.data.shape

    def test_4d_data_as_lazy(self):
        data = np.random.random((4, 10, 15))
        s = PolarDiffraction2D(data)
        s_lazy = s.as_lazy()
        assert s_lazy.__class__ == LazyPolarDiffraction2D
        assert data.shape == s_lazy.data.shape


class TestCorrelations:
    @pytest.fixture
    def flat_pattern(self):
        pd = PolarDiffraction2D(data=np.ones(shape=(2, 2, 5, 5)))
        pd.axes_manager.signal_axes[0].scale = 0.5
        pd.axes_manager.signal_axes[0].name = "theta"
        pd.axes_manager.signal_axes[1].scale = 2
        pd.axes_manager.signal_axes[1].name = "k"
        return pd

    def test_correlation_signal(self, flat_pattern):
        ac = flat_pattern.get_angular_correlation()
        assert isinstance(ac, Correlation2D)

    def test_axes_transfer(self, flat_pattern):
        ac = flat_pattern.get_angular_correlation()
        assert (
            ac.axes_manager.signal_axes[0].scale
            == flat_pattern.axes_manager.signal_axes[0].scale
        )
        assert (
            ac.axes_manager.signal_axes[1].scale
            == flat_pattern.axes_manager.signal_axes[1].scale
        )
        assert (
            ac.axes_manager.signal_axes[1].name
            == flat_pattern.axes_manager.signal_axes[1].name
        )

    @pytest.mark.parametrize(
        "mask", [None, np.zeros(shape=(5, 5)), Signal2D(np.zeros(shape=(2, 2, 5, 5)))]
    )
    def test_masking_correlation(self, flat_pattern, mask):
        ap = flat_pattern.get_angular_correlation(mask=mask)
        assert isinstance(ap, Correlation2D)

    def test_correlation_inplace(self, flat_pattern):
        ac = flat_pattern.get_angular_correlation(inplace=True)
        assert ac is None
        assert isinstance(flat_pattern, Correlation2D)

    @pytest.mark.parametrize(
        "mask", [None, np.zeros(shape=(5, 5)), Signal2D(np.zeros(shape=(2, 2, 5, 5)))]
    )
    def test_masking_angular_power(self, flat_pattern, mask):
        ap = flat_pattern.get_angular_power(mask=mask)
        print(ap)
        assert isinstance(ap, Power2D)

    def test_axes_transfer_power(self, flat_pattern):
        ac = flat_pattern.get_angular_power()
        assert ac.axes_manager.signal_axes[0].scale == 1
        assert (
            ac.axes_manager.signal_axes[1].scale
            == flat_pattern.axes_manager.signal_axes[1].scale
        )
        assert (
            ac.axes_manager.signal_axes[1].name
            == flat_pattern.axes_manager.signal_axes[1].name
        )

    def test_power_inplace(self, flat_pattern):
        ac = flat_pattern.get_angular_power(inplace=True)
        assert ac is None
        assert isinstance(flat_pattern, Power2D)


class TestDecomposition:
    def test_decomposition_is_performed(self, diffraction_pattern):
        s = PolarDiffraction2D(diffraction_pattern)
        s.decomposition()
        assert s.learning_results is not None

    def test_decomposition_class_assignment(self, diffraction_pattern):
        s = PolarDiffraction2D(diffraction_pattern)
        s.decomposition()
        assert isinstance(s, PolarDiffraction2D)



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
from pyxem.detectors.generic_flat_detector import GenericFlatDetector
from pyFAI.detectors import Detector


def test_generic_flat_detector_init():
    size_x, size_y = 256, 256
    detector = GenericFlatDetector(size_x, size_y)
    assert isinstance(detector, Detector)
    assert np.array_equal(detector.max_shape, (size_x, size_y))



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np


def calc_radius_with_distortion(x, y, xc, yc, asym, rot):
    """ calculate the distance of each 2D point from the center (xc, yc) """
    xp = x * np.cos(rot) - y * np.sin(rot)
    yp = x * np.sin(rot) + y * np.cos(rot)
    xcp = xc * np.cos(rot) - yc * np.sin(rot)
    ycp = xc * np.sin(rot) + yc * np.cos(rot)

    return np.sqrt((xp - xcp) ** 2 + asym * (yp - ycp) ** 2)


def call_ring_pattern(xcenter, ycenter):
    """
    Function to make a call to the function ring_pattern without passing the
    variables directly (necessary for using scipy.optimize.curve_fit).

    Parameters
    ----------
    xcenter : float
        The coordinate (fractional pixel units) of the diffraction
        pattern center in the first dimension
    ycenter : float
        The coordinate (fractional pixel units) of the diffraction
        pattern center in the second dimension

    Returns
    -------
    ring_pattern : function
        A function that calculates a ring pattern given a set of points and
        parameters.

    """

    def ring_pattern(
        pts, scale, amplitude, spread, direct_beam_amplitude, asymmetry, rotation
    ):
        """Calculates a polycrystalline gold diffraction pattern given a set of
        pixel coordinates (points). It uses tabulated values of the spacings
        (in reciprocal Angstroms) and relative intensities of rings derived from
        X-ray scattering factors.

        Parameters
        -----------
        pts : 1D array
            One-dimensional array of points (first half as first-dimension
            coordinates, second half as second-dimension coordinates)
        scale : float
            An initial guess for the diffraction calibration
            in 1/Angstrom units
        amplitude : float
            An initial guess for the amplitude of the polycrystalline rings
            in arbitrary units
        spread : float
            An initial guess for the spread within each ring (Gaussian width)
        direct_beam_amplitude : float
            An initial guess for the background intensity from
            the direct beam disc in arbitrary units
        asymmetry : float
            An initial guess for any elliptical asymmetry in the pattern
            (for a perfectly circular pattern asymmetry=1)
        rotation : float
            An initial guess for the rotation of the (elliptical) pattern
            in radians.

        Returns
        -------
        ring_pattern : np.array()
            A one-dimensional array of the intensities of the ring pattern
            at the supplied points.

        """
        rings = [0.4247, 0.4904, 0.6935, 0.8132, 0.8494, 0.9808, 1.0688, 1.0966]
        rings = np.multiply(rings, scale)
        amps = [1, 0.44, 0.19, 0.16, 0.04, 0.014, 0.038, 0.036]

        x = pts[: round(np.size(pts, 0) / 2)]
        y = pts[round(np.size(pts, 0) / 2) :]
        Ri = calc_radius_with_distortion(x, y, xcenter, ycenter, asymmetry, rotation)

        v = []
        denom = 2 * spread ** 2
        v.append(direct_beam_amplitude * Ri ** -2)  # np.exp((-1*(Ri)*(Ri))/d0)
        for i in [0, 1, 2, 3, 4, 5, 6, 7]:
            v.append(amps[i] * np.exp((-1 * (Ri - rings[i]) * (Ri - rings[i])) / denom))

        return (
            amplitude
            * (v[0] + v[1] + v[2] + v[3] + v[4] + v[5] + v[6] + v[7] + v[8]).ravel()
        )

    return ring_pattern


def generate_ring_pattern(
    image_size,
    mask=False,
    mask_radius=10,
    scale=100,
    amplitude=1000,
    spread=2,
    direct_beam_amplitude=500,
    asymmetry=1,
    rotation=0,
):
    """Calculate a set of rings to model a polycrystalline gold diffraction
    pattern for use in fitting for diffraction pattern calibration.
    It is suggested that the function generate_ring_pattern is used to
    find initial values (initial guess) for the parameters used in
    the function fit_ring_pattern.

    This function is written expecting a single 2D diffraction pattern
    with equal dimensions (e.g. 256x256).

    Parameters
    ----------
    mask : bool
        Choice of whether to use mask or not (mask=True will return a
        specified circular mask setting a region around
        the direct beam to zero)
    mask_radius : int
        The radius in pixels for a mask over the direct beam disc
        (the direct beam disc within given radius will be excluded
        from the fit)
    scale : float
        An initial guess for the diffraction calibration
        in 1/Angstrom units
    image_size : int
        Size of the diffraction pattern to be generated in pixels.
    amplitude : float
        An initial guess for the amplitude of the polycrystalline rings
        in arbitrary units
    spread : float
        An initial guess for the spread within each ring (Gaussian width)
    direct_beam_amplitude : float
        An initial guess for the background intensity from the
        direct beam disc in arbitrary units
    asymmetry : float
        An initial guess for any elliptical asymmetry in the pattern
        (for a perfectly circular pattern asymmetry=1)
    rotation : float
        An initial guess for the rotation of the (elliptical) pattern
        in radians.

    Returns
    -------
    image : np.array()
        Simulated ring pattern with the same dimensions as self.data

    """
    xi = np.linspace(0, image_size - 1, image_size)
    yi = np.linspace(0, image_size - 1, image_size)
    x, y = np.meshgrid(xi, yi)

    pts = np.array([x.ravel(), y.ravel()]).ravel()
    xcenter = (image_size - 1) / 2
    ycenter = (image_size - 1) / 2

    ring_pattern = call_ring_pattern(xcenter, ycenter)
    generated_pattern = ring_pattern(
        pts, scale, amplitude, spread, direct_beam_amplitude, asymmetry, rotation
    )
    generated_pattern = np.reshape(generated_pattern, (image_size, image_size))

    if mask == True:
        maskROI = calc_radius_with_distortion(
            x, y, (image_size - 1) / 2, (image_size - 1) / 2, 1, 0
        )
        maskROI[maskROI > mask_radius] = 0
        generated_pattern[maskROI > 0] *= 0

    return generated_pattern



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from hyperspy.signals import BaseSignal, Signal1D
from hyperspy.api import markers

import matplotlib.pyplot as plt
from matplotlib.cm import get_cmap
from scipy.spatial import distance_matrix
from sklearn.cluster import DBSCAN

from warnings import warn

from pyxem.signals import (
    transfer_navigation_axes,
    transfer_navigation_axes_to_signal_axes,
)
from pyxem.utils.vector_utils import detector_to_fourier
from pyxem.utils.vector_utils import calculate_norms, calculate_norms_ragged
from pyxem.utils.vector_utils import get_npeaks, filter_vectors_ragged
from pyxem.utils.vector_utils import filter_vectors_edge_ragged
from pyxem.utils.expt_utils import peaks_as_gvectors
from pyxem.utils.plot import generate_marker_inputs_from_peaks


"""
Signal class for diffraction vectors.

There are two cases that are supported:

1. A map of diffraction vectors, which will in general be a ragged signal of
signals. It the navigation dimensions of the map and contains a signal for each
peak at every position.

2. A list of diffraction vectors with dimensions < n | 2 > where n is the
number of peaks.
"""


class DiffractionVectors(BaseSignal):
    """Crystallographic mapping results containing the best matching crystal
    phase and orientation at each navigation position with associated metrics.

    Attributes
    ----------
    cartesian : np.array()
        Array of 3-vectors describing Cartesian coordinates associated with
        each diffraction vector.
    hkls : np.array()
        Array of Miller indices associated with each diffraction vector
        following indexation.
    """

    _signal_dimension = 0
    _signal_type = "diffraction_vectors"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cartesian = None
        self.hkls = None
        self.detector_shape = None
        self.pixel_calibration = None

    @classmethod
    def from_peaks(cls, peaks, center, calibration):
        """Takes a list of peak positions (pixel coordinates) and returns
        an instance of `Diffraction2D`

        Parameters
        ----------
        peaks : Signal
            Signal containing lists (np.array) of pixel coordinates specifying
            the reflection positions
        center : np.array
            Diffraction pattern center in array indices.
        calibration : np.array
            Calibration in reciprocal Angstroms per pixels for each of the dimensions.

        Returns
        -------
        vectors : :obj:`pyxem.signals.diffraction_vectors.DiffractionVectors`
            List of diffraction vectors
        """
        gvectors = peaks.map(
            peaks_as_gvectors, center=center, calibration=calibration, inplace=False
        )

        vectors = cls(gvectors)
        vectors.axes_manager.set_signal_dimension(0)

        return vectors

    def plot_diffraction_vectors(
        self,
        xlim=1.0,
        ylim=1.0,
        unique_vectors=None,
        distance_threshold=0.01,
        method="distance_comparison",
        min_samples=1,
        image_to_plot_on=None,
        image_cmap="gray",
        plot_label_colors=False,
        distance_threshold_all=0.005,
    ):  # pragma: no cover
        """Plot the unique diffraction vectors.

        Parameters
        ----------
        xlim : float
            The maximum x coordinate to be plotted.
        ylim : float
            The maximum y coordinate in reciprocal Angstroms to be plotted.
        unique_vectors : DiffractionVectors, optional
            The unique vectors to be plotted (optional). If not given, the
            unique vectors will be found by get_unique_vectors.
        distance_threshold : float, optional
            The minimum distance in reciprocal Angstroms between diffraction
            vectors for them to be considered unique diffraction vectors.
            Will be passed to get_unique_vectors if no unique vectors are
            given.
        method : str
            The method to use to determine unique vectors, if not given.
            Valid methods are 'strict', 'distance_comparison' and 'DBSCAN'.
            'strict' returns all vectors that are strictly unique and
            corresponds to distance_threshold=0.
            'distance_comparison' checks the distance between vectors to
            determine if some should belong to the same unique vector,
            and if so, the unique vector is iteratively updated to the
            average value.
            'DBSCAN' relies on the DBSCAN [1] clustering algorithm, and
            uses the Eucledian distance metric.
        min_samples : int, optional
            The minimum number of not identical vectors within one cluster
            for it to be considered a core sample, i.e. to not be considered
            noise. Will be passed to get_unique_vectors if no unique vectors
            are given. Only used if method=='DBSCAN'.
        image_to_plot_on : BaseSignal, optional
            If provided, the vectors will be plotted on top of this image.
            The image must be calibrated in terms of offset and scale.
        image_cmap : str, optional
            The colormap to plot the image in.
        plot_label_colors : bool, optional
            If True (default is False), also the vectors contained within each
            cluster will be plotted, with colors according to their
            cluster membership. If True, the unique vectors will be
            calculated by get_unique_vectors. Requires on method=='DBSCAN'.
        distance_threshold_all : float, optional
            The minimum distance, in calibrated units, between diffraction
            vectors inside one cluster for them to be plotted. Only used if
            plot_label_colors is True and requires method=='DBSCAN'.

        Returns
        -------
        fig : matplotlib figure
            The plot as a matplotlib figure.

        """
        fig = plt.figure()
        ax = fig.add_subplot(111)
        offset, scale = 0.0, 1.0
        if image_to_plot_on is not None:
            offset = image_to_plot_on.axes_manager[-1].offset
            scale = image_to_plot_on.axes_manager[-1].scale
            ax.imshow(image_to_plot_on, cmap=image_cmap)
        else:
            ax.set_xlim(-xlim, xlim)
            ax.set_ylim(ylim, -ylim)
            ax.set_aspect("equal")

        if plot_label_colors is True and method == "DBSCAN":
            clusters = self.get_unique_vectors(
                distance_threshold,
                method="DBSCAN",
                min_samples=min_samples,
                return_clusters=True,
            )[1]
            labs = clusters.labels_[clusters.core_sample_indices_]
            # Get all vectors from the clustering not considered noise
            cores = clusters.components_
            if cores.size == 0:
                warn(
                    "No clusters were found. Check parameters, or "
                    "use plot_label_colors=False."
                )
            else:
                peaks = DiffractionVectors(cores)
                peaks.axes_manager.set_signal_dimension(1)
                # Since this original number of vectors can be huge, we
                # find a reduced number of vectors that should be plotted, by
                # running a new clustering on all the vectors not considered
                # noise, considering distance_threshold_all.
                peaks = peaks.get_unique_vectors(
                    distance_threshold_all, min_samples=1, return_clusters=False
                )
                peaks_all_len = peaks.data.shape[0]
                labels_to_plot = np.zeros(peaks_all_len)
                peaks_to_plot = np.zeros((peaks_all_len, 2))
                # Find the labels of each of the peaks to plot by referring back
                # to the list of labels for the original vectors.
                for n, peak in zip(np.arange(peaks_all_len), peaks):
                    index = distance_matrix([peak.data], cores).argmin()
                    peaks_to_plot[n] = cores[index]
                    labels_to_plot[n] = labs[index]
                # Assign a color value to each label, and shuffle these so that
                # adjacent clusters hopefully get distinct colors.
                cmap_lab = get_cmap("gist_rainbow")
                lab_values_shuffled = np.arange(np.max(labels_to_plot) + 1)
                np.random.shuffle(lab_values_shuffled)
                labels_steps = np.array(
                    list(map(lambda n: lab_values_shuffled[int(n)], labels_to_plot))
                )
                labels_steps = labels_steps / (np.max(labels_to_plot) + 1)
                # Plot all peaks
                for lab, peak in zip(labels_steps, peaks_to_plot):
                    ax.plot(
                        (peak[0] - offset) / scale,
                        (peak[1] - offset) / scale,
                        ".",
                        color=cmap_lab(lab),
                    )
        if unique_vectors is None:
            unique_vectors = self.get_unique_vectors(
                distance_threshold, method=method, min_samples=min_samples
            )
        # Plot the unique vectors
        ax.plot(
            (unique_vectors.data.T[0] - offset) / scale,
            (unique_vectors.data.T[1] - offset) / scale,
            "kx",
        )
        plt.tight_layout()
        plt.axis("off")
        return fig

    def plot_diffraction_vectors_on_signal(self, signal, *args, **kwargs):
        """Plot the diffraction vectors on a signal.

        Parameters
        ----------
        signal : ElectronDiffraction2D
            The ElectronDiffraction2D signal object on which to plot the peaks.
            This signal must have the same navigation dimensions as the peaks.
        *args :
            Arguments passed to signal.plot()
        **kwargs :
            Keyword arguments passed to signal.plot()
        """
        mmx, mmy = generate_marker_inputs_from_peaks(self)
        signal.plot(*args, **kwargs)
        for mx, my in zip(mmx, mmy):
            m = markers.point(x=mx, y=my, color="red", marker="x")
            signal.add_marker(m, plot_marker=True, permanent=False)

    def get_magnitudes(self, *args, **kwargs):
        """Calculate the magnitude of diffraction vectors.

        Parameters
        ----------
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to map().

        Returns
        -------
        magnitudes : BaseSignal
            A signal with navigation dimensions as the original diffraction
            vectors containging an array of gvector magnitudes at each
            navigation position.

        """
        # If ragged the signal axes will not be defined
        if len(self.axes_manager.signal_axes) == 0:
            magnitudes = self.map(
                calculate_norms_ragged, inplace=False, *args, **kwargs
            )
        # Otherwise easier to calculate.
        else:
            magnitudes = BaseSignal(calculate_norms(self))
            magnitudes.axes_manager.set_signal_dimension(0)

        return magnitudes

    def get_magnitude_histogram(self, bins, *args, **kwargs):
        """Obtain a histogram of gvector magnitudes.

        Parameters
        ----------
        bins : numpy array
            The bins to be used to generate the histogram.
        *args:
            Arguments to get_magnitudes().
        **kwargs:
            Keyword arguments to get_magnitudes().

        Returns
        -------
        ghis : Signal1D
            Histogram of gvector magnitudes.

        """
        gmags = self.get_magnitudes(*args, **kwargs)

        if len(self.axes_manager.signal_axes) == 0:
            glist = []
            for i in gmags._iterate_signal():
                for j in np.arange(len(i[0])):
                    glist.append(i[0][j])
            gs = np.asarray(glist)
            gsig = Signal1D(gs)
            ghis = gsig.get_histogram(bins=bins)

        else:
            ghis = gmags.get_histogram(bins=bins)

        ghis.axes_manager.signal_axes[0].name = "k"
        ghis.axes_manager.signal_axes[0].units = "$A^{-1}$"

        return ghis

    def get_unique_vectors(
        self,
        distance_threshold=0.01,
        method="distance_comparison",
        min_samples=1,
        return_clusters=False,
    ):
        """Returns diffraction vectors considered unique by:
        strict comparison, distance comparison with a specified
        threshold, or by clustering using DBSCAN [1].

        Parameters
        ----------
        distance_threshold : float
            The minimum distance between diffraction vectors for them to
            be considered unique diffraction vectors. If
            distance_threshold==0, the unique vectors will be determined
            by strict comparison.
        method : str
            The method to use to determine unique vectors. Valid methods
            are 'strict', 'distance_comparison' and 'DBSCAN'.
            'strict' returns all vectors that are strictly unique and
            corresponds to distance_threshold=0.
            'distance_comparison' checks the distance between vectors to
            determine if some should belong to the same unique vector,
            and if so, the unique vector is iteratively updated to the
            average value.
            'DBSCAN' relies on the DBSCAN [1] clustering algorithm, and
            uses the Eucledian distance metric.
        min_samples : int, optional
            The minimum number of not strictly identical vectors within
            one cluster for the cluster to be considered a core sample,
            i.e. to not be considered noise. Only used for method='DBSCAN'.
        return_clusters : bool, optional
            If True (False is default), the DBSCAN clustering result is
            returned. Only used for method='DBSCAN'.

        References
        ----------
        [1] https://scikit-learn.org/stable/modules/generated/sklearn.
            cluster.DBSCAN.html

        Returns
        -------
        unique_peaks : DiffractionVectors
            The unique diffraction vectors.
        clusters : DBSCAN
            The results from the clustering, given as class DBSCAN.
            Only returned if method='DBSCAN' and return_clusters=True.
        """
        # Flatten the array of peaks to reach dimension (n, 2), where n
        # is the number of peaks.
        peaks_all = np.concatenate([peaks.ravel() for peaks in self.data.flat]).reshape(
            -1, 2
        )

        # A distance_threshold of 0 implies a strict comparison. So in that
        # case, a warning is raised unless the specified method is 'strict'.
        if distance_threshold == 0:
            if method != "strict":
                warn(
                    "distance_threshold=0 was given, and therefore "
                    "a strict comparison is used, even though the "
                    "specified method was {}".format(method)
                )
                method = "strict"

        if method == "strict":
            unique_peaks = np.unique(peaks_all, axis=0)

        elif method == "distance_comparison":
            unique_vectors, unique_counts = np.unique(
                peaks_all, axis=0, return_counts=True
            )

            unique_peaks = np.array([[0, 0]])
            unique_peaks_counts = np.array([0])

            while unique_vectors.shape[0] > 0:
                unique_vector = unique_vectors[0]
                distances = distance_matrix(np.array([unique_vector]), unique_vectors)
                indices = np.where(distances < distance_threshold)[1]

                new_count = indices.size
                new_unique_peak = np.array(
                    [
                        np.average(
                            unique_vectors[indices],
                            weights=unique_counts[indices],
                            axis=0,
                        )
                    ]
                )

                unique_peaks = np.append(unique_peaks, new_unique_peak, axis=0)

                unique_peaks_counts = np.append(unique_peaks_counts, new_count)
                unique_vectors = np.delete(unique_vectors, indices, axis=0)
                unique_counts = np.delete(unique_counts, indices, axis=0)
            unique_peaks = np.delete(unique_peaks, [0], axis=0)

        elif method == "DBSCAN":
            # All peaks are clustered by DBSCAN so that peaks within
            # one cluster are separated by distance_threshold or less.
            unique_vectors, unique_vectors_counts = np.unique(
                peaks_all, axis=0, return_counts=True
            )
            clusters = DBSCAN(
                eps=distance_threshold, min_samples=min_samples, metric="euclidean"
            ).fit(unique_vectors, sample_weight=unique_vectors_counts)
            unique_labels, unique_labels_count = np.unique(
                clusters.labels_, return_counts=True
            )
            unique_peaks = np.zeros((unique_labels.max() + 1, 2))

            # For each cluster, a center of mass is calculated based
            # on all the peaks within the cluster, and the center of
            # mass is taken as the final unique vector position.
            for n in np.arange(unique_labels.max() + 1):
                peaks_n_temp = unique_vectors[clusters.labels_ == n]
                peaks_n_counts_temp = unique_vectors_counts[clusters.labels_ == n]
                unique_peaks[n] = np.average(
                    peaks_n_temp, weights=peaks_n_counts_temp, axis=0
                )

        # Manipulate into DiffractionVectors class
        if unique_peaks.size > 0:
            unique_peaks = DiffractionVectors(unique_peaks)
            unique_peaks.axes_manager.set_signal_dimension(1)
        if return_clusters and method == "DBSCAN":
            return unique_peaks, clusters
        else:
            return unique_peaks

    def filter_magnitude(self, min_magnitude, max_magnitude, *args, **kwargs):
        """Filter the diffraction vectors to accept only those with a magnitude
        within a user specified range.

        Parameters
        ----------
        min_magnitude : float
            Minimum allowed vector magnitude.
        max_magnitude : float
            Maximum allowed vector magnitude.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to map().

        Returns
        -------
        filtered_vectors : DiffractionVectors
            Diffraction vectors within allowed magnitude tolerances.
        """
        # If ragged the signal axes will not be defined
        if len(self.axes_manager.signal_axes) == 0:
            filtered_vectors = self.map(
                filter_vectors_ragged,
                min_magnitude=min_magnitude,
                max_magnitude=max_magnitude,
                inplace=False,
                *args,
                **kwargs
            )
            # Type assignment to DiffractionVectors for return
            filtered_vectors = DiffractionVectors(filtered_vectors)
            filtered_vectors.axes_manager.set_signal_dimension(0)
        # Otherwise easier to calculate.
        else:
            magnitudes = self.get_magnitudes()
            magnitudes.data[magnitudes.data < min_magnitude] = 0
            magnitudes.data[magnitudes.data > max_magnitude] = 0
            filtered_vectors = self.data[np.where(magnitudes)]
            # Type assignment to DiffractionVectors for return
            filtered_vectors = DiffractionVectors(filtered_vectors)
            filtered_vectors.axes_manager.set_signal_dimension(1)

        transfer_navigation_axes(filtered_vectors, self)

        return filtered_vectors

    def filter_detector_edge(self, exclude_width, *args, **kwargs):
        """Filter the diffraction vectors to accept only those not within a
        user specified proximity to the detector edge.

        Parameters
        ----------
        exclude_width : int
            The width of the region adjacent to the detector edge from which
            vectors will be excluded.
        *args:
            Arguments to be passed to map().
        **kwargs:
            Keyword arguments to map().

        Returns
        -------
        filtered_vectors : DiffractionVectors
            Diffraction vectors within allowed detector region.
        """
        x_threshold = (
            self.pixel_calibration * (self.detector_shape[0] / 2)
            - self.pixel_calibration * exclude_width
        )
        y_threshold = (
            self.pixel_calibration * (self.detector_shape[1] / 2)
            - self.pixel_calibration * exclude_width
        )
        # If ragged the signal axes will not be defined
        if len(self.axes_manager.signal_axes) == 0:
            filtered_vectors = self.map(
                filter_vectors_edge_ragged,
                x_threshold=x_threshold,
                y_threshold=y_threshold,
                inplace=False,
                *args,
                **kwargs
            )
            # Type assignment to DiffractionVectors for return
            filtered_vectors = DiffractionVectors(filtered_vectors)
            filtered_vectors.axes_manager.set_signal_dimension(0)
        # Otherwise easier to calculate.
        else:
            x_inbounds = (
                np.absolute(self.data.T[0]) < x_threshold
            )  # True if vector is good to go
            y_inbounds = np.absolute(self.data.T[1]) < y_threshold
            filtered_vectors = self.data[np.logical_and(x_inbounds, y_inbounds)]
            # Type assignment to DiffractionVectors for return
            filtered_vectors = DiffractionVectors(filtered_vectors)
            filtered_vectors.axes_manager.set_signal_dimension(1)

        transfer_navigation_axes(filtered_vectors, self)

        return filtered_vectors

    def get_diffracting_pixels_map(self, in_range=None, binary=False):
        """Map of the number of vectors at each navigation position.

        Parameters
        ----------
        in_range : tuple
            Tuple (min_magnitude, max_magnitude) the minimum and maximum
            magnitude of vectors to be used to form the map.
        binary : boolean
            If True a binary image with diffracting pixels taking value == 1 is
            returned.

        Returns
        -------
        crystim : Signal2D
            2D map of diffracting pixels.
        """
        if in_range:
            filtered = self.filter_magnitude(in_range[0], in_range[1])
            xim = filtered.map(get_npeaks, inplace=False).as_signal2D((0, 1))
        else:
            xim = self.map(get_npeaks, inplace=False).as_signal2D((0, 1))
        # Make binary if specified
        if binary is True:
            xim = xim >= 1.0
        # Set properties
        xim = transfer_navigation_axes_to_signal_axes(xim, self)
        xim.change_dtype("float")
        xim.set_signal_type("signal2d")
        xim.metadata.General.title = "Diffracting Pixels Map"

        return xim

    def calculate_cartesian_coordinates(
        self, accelerating_voltage, camera_length, *args, **kwargs
    ):
        """Get cartesian coordinates of the diffraction vectors.

        Parameters
        ----------
        accelerating_voltage : float
            The acceleration voltage with which the data was acquired.
        camera_length : float
            The camera length in meters.
        """
        # Imported here to avoid circular dependency
        from diffsims.utils.sim_utils import get_electron_wavelength

        wavelength = get_electron_wavelength(accelerating_voltage)
        self.cartesian = self.map(
            detector_to_fourier,
            wavelength=wavelength,
            camera_length=camera_length * 1e10,
            inplace=False,
            parallel=False,  # TODO: For testing
            *args,
            **kwargs
        )


class DiffractionVectors2D(DiffractionVectors):
    """Crystallographic mapping results containing the best matching crystal
    phase and orientation at each navigation position with associated metrics.

    Attributes
    ----------
    cartesian : np.array()
        Array of 3-vectors describing Cartesian coordinates associated with
        each diffraction vector.
    hkls : np.array()
        Array of Miller indices associated with each diffraction vector
        following indexation.
    """

    _signal_dimension = 2

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        if self.axes_manager.signal_dimension != 2:
            self.axes_manager.set_signal_dimension(2)



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.


def normalize_vdf(im):
    """Normalizes image intensity by dividing by maximum value.

    Parameters
    ----------
    im : np.array()
        Array of image intensities

    Returns
    -------
    imn : np.array()
        Array of normalized image intensities

    """
    imn = im / im.max()
    return imn



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

from pyxem.generators.displacement_gradient_tensor_generator import (
    get_DisplacementGradientMap,
    get_single_DisplacementGradientTensor,
)
import hyperspy.api as hs
import pytest
import numpy as np


def vector_operation(z, M):
    """
    Extracts our input format and applies operations vector by vector

    Parameters
    ----------
    z : (n,2) np.array
        input vectors
    M : (2,2) np.array
        Tranformation matrix,
    Returns
    -------
    z_transformed : (n,2) np.array
        Output vectors
    """
    v_transformed = []
    for i in np.arange(0, z.shape[1]):
        v_transformed.append(np.matmul(M, z[:, i]))
    return np.asarray(v_transformed).T


def rotation(z):
    theta = np.deg2rad(3)
    c, s = np.cos(theta), np.sin(theta)
    R = np.asarray(([c, -s], [s, c]))
    return vector_operation(z, R)


def uniform_expansion(z):
    return 1.1 * z


def stretch_in_x(z):
    M = np.asarray([[1.1, 0], [0, 1]])
    return vector_operation(z, M)


def generate_test_vectors(v):
    """
    We imagine we measured 4 sets of vectors, from 4 regions of sample, normal,
    rotated, uniform expansion and uniaxial expansion.
    """
    return np.asarray([[v, rotation(v)], [uniform_expansion(v), stretch_in_x(v)]])


def generate_strain_map(vectors):
    deformed = hs.signals.Signal2D(generate_test_vectors(vectors))
    return get_DisplacementGradientMap(deformed, vectors).get_strain_maps()


@pytest.fixture()
def xy_vectors():
    xy = np.asarray([[1, 0], [0, 1]])
    return generate_strain_map(xy)


@pytest.fixture()
def left_handed():
    not_xy = np.asarray(([1, 1], [1, -1.2]))
    return generate_strain_map(not_xy)


@pytest.fixture()
def multi_vector():
    four_vectors = np.asarray([[1, 0, 1, 1], [0, 1, -1, 1]])
    return generate_strain_map(four_vectors)


""" Each of these basis should return the same results, in an xy basis"""


def test_results_returned_correctly_in_same_basis(
    xy_vectors, left_handed, multi_vector
):
    """ Basic test of the summary statement for this section """
    np.testing.assert_almost_equal(xy_vectors.data, left_handed.data, decimal=2)
    np.testing.assert_almost_equal(xy_vectors.data, multi_vector.data, decimal=2)


def test_trivial_weight_function_case(xy_vectors):
    """ If weights are [1,1,1,1] the result should be the same as weights=None"""
    weights = [1, 1, 1, 1]
    four_vectors = np.asarray([[1, 0, 1, 1], [0, 1, -1, 1]])
    deformed = hs.signals.Signal2D(generate_test_vectors(four_vectors))
    weight_strain_map = get_DisplacementGradientMap(
        deformed, four_vectors, weights=weights
    ).get_strain_maps()
    np.testing.assert_almost_equal(xy_vectors.data, weight_strain_map.data, decimal=2)


def test_weight_function_behaviour():
    """ Confirms that  a weight function [1,1,2,2] on [a,a,b,b] gives (2a+4b)/6 as the strain"""
    multi_vector_array = np.asarray([[1, 0, 1, 1], [0, 1, -1, 1]])
    strained_by_1pc_in_x = vector_operation(
        multi_vector_array, np.asarray([[1.01, 0], [0, 1]])
    )  # first  2
    strained_by_2pc_in_x = vector_operation(
        multi_vector_array, np.asarray([[1.02, 0], [0, 1]])
    )  # second 2
    weights = [1, 1, 2, 2]  # ((0.1*2 + 0.2*4)/6) = 0.166666
    vectors = np.concatenate(
        (strained_by_1pc_in_x[:, :2], strained_by_2pc_in_x[:, 2:]), axis=1
    )
    deformed = hs.signals.Signal2D(np.asarray([[vectors, vectors], [vectors, vectors]]))
    strain_map = get_DisplacementGradientMap(
        deformed, multi_vector_array, weights=weights
    ).get_strain_maps()
    np.testing.assert_almost_equal(
        strain_map.inav[0].isig[0, 0].data[0], -1.0166666 + 1, decimal=2
    )



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from pyxem.utils.subpixel_refinements_utils import get_experimental_square
from pyxem.utils.subpixel_refinements_utils import get_simulated_disc

from skimage import draw


@pytest.fixture()
def exp_disc():
    ss, disc_radius, upsample_factor = int(60), 6, 10

    arr = np.zeros((ss, ss))
    rr, cc = draw.circle(
        int(ss / 2) + 20, int(ss / 2) - 10, radius=disc_radius, shape=arr.shape
    )
    arr[rr, cc] = 1
    return arr


@pytest.mark.filterwarnings("ignore::UserWarning")  # various skimage warnings
def test_experimental_square_size(exp_disc):
    square = get_experimental_square(exp_disc, [17, 19], 6)
    assert square.shape[0] == int(6)
    assert square.shape[1] == int(6)


def test_failure_for_non_even_entry_to_get_simulated_disc():
    with pytest.raises(ValueError, match="'square_size' must be an even number"):
        disc = get_simulated_disc(61, 5)


def test_failure_for_non_even_errors_get_experimental_square(exp_disc):
    with pytest.raises(ValueError, match="'square_size' must be an even number"):
        square = get_experimental_square(exp_disc, [17, 19], 7)



name = "pyxem"
version = "0.11.0"
author = "Duncan Johnstone, Phillip Crout"
copyright = "copyright 2016-2020, the pyxem developers"
credits = [
    "Duncan Johnstone",
    "Phillip Crout",
    "Joonatan Laulainen",
    "Simon Høgås",
    "Ben Martineau",
    "Tina Bergh",
    "Stef Smeets",
    "Carter Francis",
    "Eirik Opheim",
    "Eric Prestat",
    "Sean Collins",
    "Tom Furnival",
    "Mohsen Danaie",
    "Håkon Wiik Ånes",
    "Tiarnan Doherty",
    "Affan Iqbal",
    "Jedrzej Morzy",
    "Tomas Ostasevicius",
    "Rob Tovey",
    "Matt von Lany",
    "Tim Poon",
    "Endre Jacobsen",
]
license = "GPLv3"
maintainer = "Duncan Johnstone, Phillip Crout"
email = "pyxem.team@gmail.com"
status = "development"



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

# a lot of stuff depends on this, so we have to create it first

import os

import numpy as np
import dask.array as da
from math import floor
from scipy.signal import find_peaks
import h5py
from struct import unpack

from pyxem.signals.electron_diffraction2d import LazyElectronDiffraction2D


def load_mib(mib_path, reshape=True, flip=True):
    """Read a .mib file or an h5 stack file using dask and return as a lazy pyXem / hyperspy signal.

    Parameters
    ----------
    mib_path : str
        The full path of the .mib file to be read.
    reshape: boolean
        Keyword argument to control reshaping of the stack (default is True).
        It attepmts to reshape using the flyback pixel.
    flip: boolean
        Keyword argument to vertically flip the diffraction signal (default)
        or return unchanged. The metadata is updated accordingly.

    Returns
    -------
    data_pxm : pyxem.signals.LazyElectronDiffraction2D
    If the data is detected to be STEM is reshaped using two functions, one using the
    exposure times appearing on the header and if no exposure times available using the
    sum frames and detecting the flyback frames. If TEM data, a single frame or if
    reshaping the STEM fails, the stack is returned.
                The metadata adds the following domains for STEM mib file:
                General
                │   └── title =
                └── Signal
                    ├── binned = False
                    ├── exposure_time = 0.001
                    ├── flip = True
                    ├── flyback_times = [0.066, 0.071, 0.065, 0.017825]
                    ├── frames_number_skipped = 90
                    ├── scan_X = 256
                    └── signal_type = STEM
                The returned metadata for TEM mib file:
                General
                │   └── title =
                └── Signal
                    ├── binned = False
                    ├── exposure_time = [0.0001]
                    ├── flyback_times = None
                    ├── frames_number_skipped = None
                    ├── scan_X = None
                    └── signal_type = TEM
    """
    hdr_stuff = _parse_hdr(mib_path)
    width = hdr_stuff["width"]
    height = hdr_stuff["height"]
    width_height = width * height

    data = _mib_to_daskarr(mib_path)
    depth = _get_mib_depth(hdr_stuff, mib_path)
    hdr_bits = _get_hdr_bits(hdr_stuff)
    if hdr_stuff["Counter Depth (number)"] == 1:
        # RAW 1 bit data: the header bits are written as uint8 but the frames
        # are binary and need to be unpacked as such.
        data = data.reshape(-1, int(width_height / 8 + hdr_bits))
        data = data[:, hdr_bits:]
        # get the shape axis 1 before unpackbit
        s0 = data.shape[0]
        s1 = data.shape[1]
        data = np.unpackbits(data)
        data.reshape(s0, s1 * 8)
    else:
        data = data.reshape(-1, int(width_height + hdr_bits))
        data = data[:, hdr_bits:]
    if hdr_stuff["raw"] == "R64":
        data = _untangle_raw(data, hdr_stuff, depth)
    elif hdr_stuff["raw"] == "MIB":
        data = data.reshape(depth, width, height)

    # if small mib file read all the exposure times otherwise just the 10% default
    if os.stat(mib_path).st_size * 1e9 < 0.1:
        exp_times_list = _read_exposures(mib_path, pct_frames_to_read=1.0)
    else:
        exp_times_list = _read_exposures(mib_path)
    data_dict = _STEM_flag_dict(exp_times_list)

    if hdr_stuff["Assembly Size"] == "2x2":
        # add_crosses expects a dask array object
        data = _add_crosses(data)

    data_pxm = LazyElectronDiffraction2D(data)

    # Tranferring dict info to metadata
    if data_dict["STEM_flag"] == 1:
        data_pxm.metadata.Signal.signal_type = "STEM"
    else:
        data_pxm.metadata.Signal.signal_type = "TEM"
    data_pxm.metadata.Signal.scan_X = data_dict["scan_X"]
    data_pxm.metadata.Signal.exposure_time = data_dict["exposure time"]
    data_pxm.metadata.Signal.frames_number_skipped = data_dict[
        "number of frames_to_skip"
    ]
    data_pxm.metadata.Signal.flyback_times = data_dict["flyback_times"]
    if reshape:
        # only attempt reshaping if it is not already reshaped!
        if len(data_pxm.data.shape) == 3:
            try:
                if data_pxm.metadata.Signal.signal_type == "TEM":
                    print(
                        "This mib file appears to be TEM data. The stack is returned with no reshaping."
                    )
                    return data_pxm
                # to catch single frames:
                if data_pxm.axes_manager[0].size == 1:
                    print("This mib file is a single frame.")
                    return data_pxm
                # If the exposure time info not appearing in the header bits use reshape_4DSTEM_SumFrames
                # to reshape otherwise use reshape_4DSTEM_FlyBack function
                if (
                    data_pxm.metadata.Signal.signal_type == "STEM"
                    and data_pxm.metadata.Signal.exposure_time is None
                ):
                    print("reshaping using sum frames intensity")
                    (data_pxm, skip_ind) = reshape_4DSTEM_SumFrames(data_pxm)
                    data_pxm.metadata.Signal.signal_type = "STEM"
                    data_pxm.metadata.Signal.frames_number_skipped = skip_ind
                else:
                    print("reshaping using flyback pixel")
                    data_pxm = reshape_4DSTEM_FlyBack(data_pxm)
            except TypeError:
                print(
                    "Warning: Reshaping did not work or TEM data with no exposure info. Returning the stack with no reshaping!"
                )
                return data_pxm
            except ValueError:
                print(
                    "Warning: Reshaping did not work or TEM data with no exposure info. Returning the stack with no reshaping!"
                )
                return data_pxm
    if flip:
        data_pxm.data = np.flip(data_pxm.data, axis=2)
        data_pxm.metadata.Signal.flip = True
    else:
        data_pxm.metadata.Signal.flip = False
    return data_pxm


def mib_to_h5stack(fp, save_path, mmap_mode="r"):
    """
    Read a .mib file using memory mapping where the array
    is stored on disk and not directly loaded, but may be treated
    like a dask array. It writes the data in chunks into an h5 file.

    Parameters
    ----------
    fp: str
        Filepath of .mib file to be loaded.
    save_path: str
        h5 filename path to save the h5 file stack
    mmap_mode: str
        default 'r' - {None, 'r+', 'r', 'w+', 'c'}, optional
        If not None, then memory-map the file, using the given mode
        (see `numpy.memmap`).  The mode has no effect for pickled or
        zipped files.

    Returns
    -------
    None
    """
    # check to see if the h5 path already exists and if so raise warning
    if os.path.exists(save_path):
        print(
            "The h5 path provided already exists. Change file name to avoid overwrite."
        )
        return
    hdr_info = _parse_hdr(fp)
    width = hdr_info["width"]
    height = hdr_info["height"]

    record_by = hdr_info["record-by"]
    depth = _get_mib_depth(hdr_info, fp)

    data = _mib_to_daskarr(fp)
    hdr_bits = _get_hdr_bits(hdr_info)

    if record_by == "vector":  # spectral image
        size = (height, width, depth)
        data = data.reshape(size)

    elif record_by == "image":  # stack of images
        width_height = width * height

        # remove headers at the beginning of each frame and reshape
        if hdr_info["raw"] == "R64":
            if hdr_info["Assembly Size"] == "2x2":
                if hdr_info["Counter Depth (number)"] == 1:
                    _stack_h5dump(data, hdr_info, save_path, raw_binary=True)
                else:
                    # All the other counter depths RAW format
                    _stack_h5dump(data, hdr_info, save_path)
        elif hdr_info["raw"] == "MIB":
            _stack_h5dump(data, hdr_info, save_path)
    return


def h5stack_to_pxm(h5_path, mib_path, flip=True):
    """
    this function reads the saved stack h5 file into a reshaped pyxem.signals.LazyElectronDiffraction2D object
    chunks are defined as (100, det_x, det_y)

    Parameters
    ----------
    h5_path: str
        full path and name of the h5 stack file
    mib_path: str
        full path and name of the mib file
    flip: boolean
        Keyword argument to vertically flip the diffraction signal (default)
        or return unchanged. The metadata is updated accordingly.

    Returns
    -------
    data_pxm: pyxem.signals.LazyElectronDiffraction2D
    """
    hdr_info = _parse_hdr(mib_path)
    f = h5py.File(h5_path, "r")

    data = f["data_stack"]

    data_pxm = LazyElectronDiffraction2D(data)

    if hdr_info["Assembly Size"] == "2x2":
        data = data_pxm.data
        # add_crosses expects a dask array object
        data = _add_crosses(data)
        data_pxm = LazyElectronDiffraction2D(data)

    if os.stat(mib_path).st_size * 1e9 < 0.1:
        exp_times_list = _read_exposures(mib_path, pct_frames_to_read=1.0)
    else:
        exp_times_list = _read_exposures(mib_path)
    data_dict = _STEM_flag_dict(exp_times_list)

    # Tranferring dict info to metadata
    if data_dict["STEM_flag"] == 1:
        data_pxm.metadata.Signal.signal_type = "STEM"
    else:
        data_pxm.metadata.Signal.signal_type = "TEM"
    data_pxm.metadata.Signal.scan_X = data_dict["scan_X"]
    data_pxm.metadata.Signal.exposure_time = data_dict["exposure time"]
    data_pxm.metadata.Signal.frames_number_skipped = data_dict[
        "number of frames_to_skip"
    ]
    data_pxm.metadata.Signal.flyback_times = data_dict["flyback_times"]

    if (
        data_pxm.metadata.Signal.signal_type == "TEM"
        and data_pxm.metadata.Signal.exposure_time != None
    ):
        print(
            "This mib file appears to be TEM data. The stack is returned with no reshaping."
        )

    # to catch single frames:
    if data_pxm.axes_manager[0].size == 1:
        print("This mib file is a single frame.")

    try:
        # If the exposure time info not appearing in the header bits use reshape_4DSTEM_SumFrames
        # to reshape otherwise use reshape_4DSTEM_FlyBack function
        if data_pxm.metadata.Signal.exposure_time is None:
            (data_pxm, skip_ind) = reshape_4DSTEM_SumFrames(data_pxm)
            data_pxm.metadata.Signal.signal_type = "STEM"
            data_pxm.metadata.Signal.frames_number_skipped = skip_ind
        else:
            print("reshaping using flyback pixel")
            data_pxm = reshape_4DSTEM_FlyBack(data_pxm)
    except TypeError:
        print(
            "Warning: Reshaping did not work or TEM data with no exposure info. Returning the stack with no reshaping!"
        )
    except ValueError:
        print(
            "Warning: Reshaping did not work or TEM data with no exposure info. Returning the stack with no reshaping!"
        )
    if flip:
        data_pxm.data = np.flip(data_pxm.data, axis=2)
        data_pxm.metadata.Signal.flip = True
    else:
        data_pxm.metadata.Signal.flip = False
    return data_pxm


def _manageHeader(fname):
    """Get necessary information from the header of the .mib file.

    Parameters
    ----------
    fname : str
        Filename for header file.

    Returns
    -------
    hdr : tuple
        (DataOffset,NChips,PixelDepthInFile,sensorLayout,Timestamp,shuttertime,bitdepth)

    Examples
    --------
    #Output for 6bit 256*256 data:
    #(768, 4, 'R64', '2x2', '2019-06-14 11:46:12.607836', 0.0002, 6)
    #Output for 12bit single frame nor RAW:
    #(768, 4, 'U16', '2x2', '2019-06-06 11:12:42.001309', 0.001, 12)

    """
    Header = str()
    with open(fname, "rb") as input:
        aByte = input.read(1)
        Header += str(aByte.decode("ascii"))
        # This gets rid of the header
        while aByte and ord(aByte) != 0:

            aByte = input.read(1)
            Header += str(aByte.decode("ascii"))

    elements_in_header = Header.split(",")

    DataOffset = int(elements_in_header[2])

    NChips = int(elements_in_header[3])

    PixelDepthInFile = elements_in_header[6]
    sensorLayout = elements_in_header[7].strip()
    Timestamp = elements_in_header[9]
    shuttertime = float(elements_in_header[10])

    if PixelDepthInFile == "R64":
        bitdepth = int(elements_in_header[18])  # RAW
    elif PixelDepthInFile == "U16":
        bitdepth = 12
    elif PixelDepthInFile == "U08":
        bitdepth = 6
    elif PixelDepthInFile == "U32":
        bitdepth = 24

    hdr = (
        DataOffset,
        NChips,
        PixelDepthInFile,
        sensorLayout,
        Timestamp,
        shuttertime,
        bitdepth,
    )

    return hdr


def _parse_hdr(fp):
    """Parse information from mib file header info from _manageHeader function.

    Parameters
    ----------
    fp : str
        Filepath to .mib file.

    Returns
    -------
    hdr_info : dict
        Dictionary containing header info extracted from .mib file.
        the entries of the dictionary are as follows:
        'width': int
            pixels, detector number of pixels in x direction,
        'height': int
            pixels detector number of pixels in y direction,
        'Assembly Size': str
            configuration of the detector chips, e.g. '2x2' for quad,
        'offset': int
            number of characters in the header before the first frame starts,
        'data-type': str
            always 'unsigned',
        'data-length': str
            identifying dtype,
        'Counter Depth (number)': int
            counter bit depth,
        'raw': str
            regular binary 'MIB' or raw binary 'R64',
        'byte-order': str
            always 'dont-care',
        'record-by': str
            'image' or 'vector' - only 'image' encountered,
        'title': str
            path of the mib file without extension, e.g. '/dls/e02/data/2020/cm26481-1/Merlin/testing/20200204 115306/test',
        'date': str
            date created, e.g. '20200204',
        'time': str
            time created, e.g. '11:53:32.295336',
        'data offset': int
            number of characters at the header.
    """
    hdr_info = {}

    read_hdr = _manageHeader(fp)

    # Set the array size of the chip

    if read_hdr[3] == "1x1":
        hdr_info["width"] = 256
        hdr_info["height"] = 256
    elif read_hdr[3] == "2x2":
        hdr_info["width"] = 512
        hdr_info["height"] = 512

    hdr_info["Assembly Size"] = read_hdr[3]

    # Set mib offset
    hdr_info["offset"] = read_hdr[0]
    # Set data-type
    hdr_info["data-type"] = "unsigned"
    # Set data-length
    if read_hdr[6] == "1":
        # Binary data recorded as 8 bit numbers
        hdr_info["data-length"] = "8"
    else:
        # Changes 6 to 8 , 12 to 16 and 24 to 32 bit
        cd_int = int(read_hdr[6])
        hdr_info["data-length"] = str(int((cd_int + cd_int / 3)))

    hdr_info["Counter Depth (number)"] = int(read_hdr[6])
    if read_hdr[2] == "R64":
        hdr_info["raw"] = "R64"
    else:
        hdr_info["raw"] = "MIB"
    # Set byte order
    hdr_info["byte-order"] = "dont-care"
    # Set record by to stack of images
    hdr_info["record-by"] = "image"

    # Set title to file name
    hdr_info["title"] = fp.split(".")[0]
    # Set time and date
    # Adding the try argument to accommodate the new hdr formatting as of April 2018
    try:
        year, month, day_time = read_hdr[4].split("-")
        day, time = day_time.split(" ")
        hdr_info["date"] = year + month + day
        hdr_info["time"] = time
    except BaseException:
        day, month, year_time = read_hdr[4].split("/")
        year, time = year_time.split(" ")
        hdr_info["date"] = year + month + day
        hdr_info["time"] = time

    hdr_info["data offset"] = read_hdr[0]

    return hdr_info


def _add_crosses(a):
    """
    Adds 3 pixel buffer cross to quad chip data.

    Parameters
    ----------
    a : dask.array
        Stack of raw frames or reshaped dask array object, prior to dimension reshaping, to insert
        3 pixel buffer cross into.

    Returns
    -------
    b : dask.array
        Stack of frames or reshaped 4DSTEM object including 3 pixel buffer cross in the diffraction plane.
    """
    original_shape = a.shape

    if len(original_shape) == 4:
        a = a.reshape(
            original_shape[0] * original_shape[1], original_shape[2], original_shape[3]
        )

    a_half = int(original_shape[-1] / 2), int(original_shape[-2] / 2)
    # Define 3 pixel wide cross of zeros to pad raw data
    if len(original_shape) == 4:
        z_array = da.zeros(
            (original_shape[0] * original_shape[1], original_shape[-2], 3),
            dtype=a.dtype,
        )
        z_array2 = da.zeros(
            (original_shape[0] * original_shape[1], 3, original_shape[-1] + 3),
            dtype=a.dtype,
        )
    else:
        z_array = da.zeros((original_shape[0], original_shape[-2], 3), dtype=a.dtype)
        z_array2 = da.zeros(
            (original_shape[0], 3, original_shape[-1] + 3), dtype=a.dtype
        )

    # Insert blank cross into raw data
    b = da.concatenate((a[:, :, : a_half[1]], z_array, a[:, :, a_half[1] :]), axis=-1)

    b = da.concatenate((b[:, : a_half[0], :], z_array2, b[:, a_half[0] :, :]), axis=-2)

    if len(original_shape) == 4:
        b = b.reshape(
            original_shape[0],
            original_shape[1],
            original_shape[2] + 3,
            original_shape[3] + 3,
        )

    return b


def _get_mib_depth(hdr_info, fp):
    """Determine the total number of frames based on .mib file size.

    Parameters
    ----------
    hdr_info : dict
        Dictionary containing header info extracted from .mib file.
    fp : filepath
        Path to .mib file.

    Returns
    -------
    depth : int
        Number of frames in the stack
    """
    # Define standard frame sizes for quad and single medipix chips
    if hdr_info["Assembly Size"] == "2x2":
        mib_file_size_dict = {
            "1": 33536,
            "6": 262912,
            "12": 525056,
            "24": 1049344,
        }
    if hdr_info["Assembly Size"] == "1x1":
        mib_file_size_dict = {
            "1": 8576,
            "6": 65920,
            "12": 131456,
            "24": 262528,
        }

    file_size = os.path.getsize(fp[:-3] + "mib")
    if hdr_info["raw"] == "R64":

        single_frame = mib_file_size_dict.get(str(hdr_info["Counter Depth (number)"]))
        depth = int(file_size / single_frame)
    elif hdr_info["raw"] == "MIB":
        if hdr_info["Counter Depth (number)"] == "1":
            # 1 bit and 6 bit non-raw frames have the same size
            single_frame = mib_file_size_dict.get("6")
            depth = int(file_size / single_frame)
        else:
            single_frame = mib_file_size_dict.get(
                str(hdr_info["Counter Depth (number)"])
            )
            depth = int(file_size / single_frame)

    return depth


def _mib_to_daskarr(fp, mmap_mode="r"):
    """
    Reads the binary mib file into a numpy memmap object and returns as dask array object

    Parameters
    --------------
    fp: str
        MIB file name / path
    mmap_mode: str
        memmpap read mode - default is 'r'
    Returns
    --------------
    data_da: dask array
        data as a dask array object
    """
    hdr_info = _parse_hdr(fp)
    data_length = hdr_info["data-length"]
    data_type = hdr_info["data-type"]
    endian = hdr_info["byte-order"]
    read_offset = 0

    if data_type == "signed":
        data_type = "int"
    elif data_type == "unsigned":
        data_type = "uint"
    elif data_type == "float":
        pass
    else:
        raise TypeError('Unknown "data-type" string.')

    # mib data always big-endian
    endian = ">"
    data_type += str(int(data_length))
    # uint1 not a valid dtype
    if data_type == "uint1":
        data_type = "uint8"
        data_type = np.dtype(data_type)
    else:
        data_type = np.dtype(data_type)
    data_type = data_type.newbyteorder(endian)

    data_mem = np.memmap(fp, offset=read_offset, dtype=data_type, mode=mmap_mode)
    data_da = da.from_array(data_mem, chunks='auto')
    return data_da


def _get_hdr_bits(hdr_info):
    """
    gets the number of character bits for the header for each frame given the data type
    Parameters
    ----------
    hdr_info: dict
        output of the parse_hdr function

    Returns
    -------
    hdr_bits: int
        number of characters in the header
    """
    data_length = hdr_info["data-length"]
    data_type = hdr_info["data-type"]

    if data_type == "signed":
        data_type = "int"
    elif data_type == "unsigned":
        data_type = "uint"
    elif data_type == "float":
        pass
    else:
        raise TypeError('Unknown "data-type" string.')

    # mib data always big-endian
    endian = ">"
    data_type += str(int(data_length))
    # uint1 not a valid dtype
    if data_type == "uint1":
        data_type = "uint8"
        data_type = np.dtype(data_type)
    else:
        data_type = np.dtype(data_type)
    data_type = data_type.newbyteorder(endian)

    if data_length == "1":
        hdr_multiplier = 1
    else:
        hdr_multiplier = (int(data_length) / 8) ** -1

    hdr_bits = int(hdr_info["data offset"] * hdr_multiplier)

    return hdr_bits


def _read_exposures(fp, pct_frames_to_read=0.1):
    """
    Looks into the frame times of the first frames to see if they are all the same (TEM) or there is a more intense
    flyback (4D-STEM). This works due to the way we trigger the 4DSTEM acquisitions at ePSIC.
    For this to work, the tick in the Merlin software to print exp time into header must be selected!

    Parameters
    -------------
    fp: str
        MIB file name / path
    pct_frames_to_read : float
        Percentage of frames to read, default value 0.1

    Returns
    ------------
    exp_time: list
        List of frame exposure times in seconds
    """
    hdr_info = _parse_hdr(fp)
    width = hdr_info["width"]
    height = hdr_info["height"]
    depth = _get_mib_depth(hdr_info, fp)

    record_by = hdr_info["record-by"]

    data = _mib_to_daskarr(fp)
    hdr_bits = _get_hdr_bits(hdr_info)

    if record_by == "vector":  # spectral image
        size = (height, width, depth)
        data = data.reshape(size)
    elif record_by == "image":  # stack of images
        width_height = width * height
        # remove headers at the beginning of each frame and reshape

        if hdr_info["raw"] == "R64":
            try:
                # the header for the case of 12 bit data should be unpacked first
                if hdr_info["Counter Depth (number)"] == 12:
                    data = data.reshape(-1, width_height + hdr_bits)[:, :68]
                    data_crop = data[: int(depth * pct_frames_to_read)]
                    d = data_crop.compute()
                    exp_time = []
                    for frame in range(d.shape[0]):
                        frame_text = str()
                        for item in d[frame]:
                            temp = unpack("cc", item)
                            c1 = temp[1].decode("ascii")
                            c2 = temp[0].decode("ascii")
                            frame_text = frame_text + c1
                            frame_text = frame_text + c2
                        exp_time.append(float(frame_text[71:79]))
                else:
                    if hdr_info["Counter Depth (number)"] == 1:
                        # RAW 1 bit data: the header bits are written as uint8 but the frames
                        # are binary and need to be unpacked as such.
                        data = data.reshape(-1, width_height / 8 + hdr_bits)[:, 71:79]

                    else:
                        data = data.reshape(-1, width_height + hdr_bits)[:, 71:79]
                    data = data[
                        :,
                    ]
                    data_crop = data[: int(depth * pct_frames_to_read)]
                    d = data_crop.compute()
                    exp_time = []
                    for line in range(d.shape[0]):
                        str_list = [chr(d[line][n]) for n in range(d.shape[1])]
                        exp_time.append(float("".join(str_list)))
            except ValueError:
                print("Frame exposure times are not appearing in header!")
        elif hdr_info["raw"] == "MIB":
            try:
                # the header for the case of 12 bit data should be unpacked first
                if hdr_info["Counter Depth (number)"] == 12:
                    data = data.reshape(-1, width_height + hdr_bits)[:, :68]
                    data_crop = data[: int(depth * pct_frames_to_read)]
                    d = data_crop.compute()
                    exp_time = []
                    for frame in range(d.shape[0]):
                        frame_text = str()
                        for item in d[frame]:
                            temp = unpack("cc", item)
                            c1 = temp[1].decode("ascii")
                            c2 = temp[0].decode("ascii")
                            frame_text = frame_text + c1
                            frame_text = frame_text + c2
                        exp_time.append(float(frame_text[71:79]))
                # all the other cases are 8 bit
                else:
                    data = data.reshape(-1, width_height + hdr_bits)[:, 71:79]
                    data = data[
                        :,
                    ]
                    data_crop = data[: int(depth * pct_frames_to_read)]
                    d = data_crop.compute()
                    exp_time = []
                    for line in range(d.shape[0]):
                        str_list = [chr(d[line][n]) for n in range(d.shape[1])]
                        exp_time.append(float("".join(str_list)))
            except ValueError:
                print("Frame exposure times are not appearing in header!")
    return exp_time


def _STEM_flag_dict(exp_times_list):
    """
    Determines whether a .mib file contains STEM or TEM data and how many
    frames to skip due to triggering from a list of exposure times.

    Parameters
    ----------
    exp_times_list : list
        List of exposure times extracted from a .mib file.

    Returns
    -------
    output : dict
        Dictionary containing the following entries:
        'STEM_flag': boolean
            1 if STEM data detected,
         'scan_X': int
            number of probe positions in a line,
         'exposure time': float
            exposure time detected per frame in seconds,
         'number of frames_to_skip': int
            number of frames to skip at the beginning before reshaping,
         'flyback_times': list of floats
            list of detected overexposed flyback frames as list
    """
    output = {}
    times_set = set(exp_times_list)
    # If single exposure times in header, treat as TEM data.
    if len(times_set) == 1:
        output["STEM_flag"] = 0
        output["scan_X"] = None
        output["exposure time"] = list(times_set)
        output["number of frames_to_skip"] = None
        output["flyback_times"] = None
    # In case exp times not appearing in header treat as TEM data
    elif len(times_set) == 0:

        output["STEM_flag"] = 0
        output["scan_X"] = None
        output["exposure time"] = None
        output["number of frames_to_skip"] = None
        output["flyback_times"] = None
    # Otherwise, treat as STEM data.
    else:
        STEM_flag = 1
        # Check that the smallest time is the majority of the values
        exp_time = max(times_set, key=exp_times_list.count)
        if exp_times_list.count(exp_time) < int(0.9 * len(exp_times_list)):
            print("Something has gone wrong with the triggering!")
        peaks = [i for i, e in enumerate(exp_times_list) if e != exp_time]
        # Diff between consecutive elements of the array
        lines = np.ediff1d(peaks)

        if len(set(lines)) == 1:
            scan_X = lines[0]
            frames_to_skip = peaks[0]
        else:
            # Assuming the last element to be the line length
            scan_X = lines[-1]
            check = np.ravel(np.where(lines == scan_X, True, False))
            # Checking line lengths
            start_ind = np.where(check == False)[0][-1] + 2
            frames_to_skip = peaks[start_ind]

        flyback_times = list(times_set)
        flyback_times.remove(exp_time)
        output["STEM_flag"] = STEM_flag
        output["scan_X"] = scan_X
        output["exposure time"] = exp_time
        output["number of frames_to_skip"] = frames_to_skip
        output["flyback_times"] = flyback_times

    return output


def _stack_h5dump(data, hdr_info, saving_path, raw_binary=False, stack_num=1000):
    """
    Incremental reading of a large stack dask array object and saving it in a h5 file.

    Parameters
    ----------
    data: dask array
    hdr_info: dict
        header info parsed by the parse_hdr function
    saving_path: str
        h5 file name and path
    raw_binary: boolean
        default False - Need to be True for binary RAW data
    stack_num: int
        number of frames written to the h5 file in each iteration. default set at 1000

    Returns
    -------
    None
    """
    hdr_bits = _get_hdr_bits(hdr_info)
    width = hdr_info["width"]
    height = hdr_info["height"]
    width_height = width * height
    if raw_binary is True:
        # RAW 1 bit data: the header bits are written as uint8 but the frames
        # are binary and need to be unpacked as such.
        data = data.reshape(-1, int(width_height / 8 + hdr_bits))
    else:
        data = data.reshape(-1, int(width_height + hdr_bits))

    data = data[:, hdr_bits:]
    iters_num = int(data.shape[0] / stack_num) + 1
    for i in range(iters_num):
        if (i + 1) * stack_num < data.shape[0]:
            if i == 0:
                data_dump0 = data[: (i + 1) * stack_num, :]
                if raw_binary is True:
                    data_dump1 = np.unpackbits(data_dump0)
                    data_dump1.reshape(data_dump0.shape[0], data_dump0.shape[1] * 8)
                    data_dump1 = _untangle_raw(
                        data_dump1, hdr_info, data_dump0.shape[0]
                    )
                else:
                    data_dump1 = _untangle_raw(
                        data_dump0, hdr_info, data_dump0.shape[0]
                    )

                _h5_chunk_write(data_dump1, saving_path)
                del data_dump0
                del data_dump1
            else:
                data_dump0 = data[i * stack_num : (i + 1) * stack_num, :]
                if raw_binary is True:
                    data_dump1 = np.unpackbits(data_dump0)
                    data_dump1.reshape(data_dump0.shape[0], data_dump0.shape[1] * 8)
                    data_dump1 = _untangle_raw(
                        data_dump1, hdr_info, data_dump0.shape[0]
                    )
                else:
                    data_dump1 = _untangle_raw(
                        data_dump0, hdr_info, data_dump0.shape[0]
                    )
                _h5_chunk_write(data_dump1, saving_path)
                del data_dump0
                del data_dump1
        else:
            data_dump0 = data[i * stack_num :, :]
            if raw_binary is True:
                data_dump1 = np.unpackbits(data_dump0)
                data_dump1.reshape(data_dump0.shape[0], data_dump0.shape[1] * 8)
                data_dump1 = _untangle_raw(data_dump1, hdr_info, data_dump0.shape[0])
            else:
                data_dump1 = _untangle_raw(data_dump0, hdr_info, data_dump0.shape[0])
            _h5_chunk_write(data_dump1, saving_path)
            del data_dump0
            del data_dump1
            return


def _h5_chunk_write(data, saving_path):
    """
    Incremental saving of the data into h5 file
    if the h5 file does not exists, creates it and if it does appends the data to the existing dataset
    h5 dataset key: 'data_stack'

    Parameters
    ----------
    data: dask array
    saving_path: str
        path and name of the h5 file

    Returns
    -------
    None
    """
    if os.path.exists(saving_path):
        print("appending to existing dataset")
        with h5py.File(saving_path, "a") as hf:
            hf["data_stack"].resize((hf["data_stack"].shape[0] + data.shape[0]), axis=0)
            hf["data_stack"][-data.shape[0] :, :, :] = data
    else:
        hf = h5py.File(saving_path, "w")
        print("creating the h5 file for the data_stack")
        hf.create_dataset(
            "data_stack",
            data=data,
            maxshape=(None, data.shape[1], data.shape[2]),
            compression="gzip",
        )
    return


def _untangle_raw(data, hdr_info, stack_size):
    """
    Corrects for the tangled raw mib format - Only the case for quad chip is considered here.

    Parameters
    --------
        data: dask array
            as stack with the detector array unreshaped, e.g. for a single frame 512*512: (1, 262144)
        hdr_info: dict
            info read from the header- ouput of the _parse_hdr function
        stack_size: int
            The number of frames in the data

    Outputs
    ----------
    untangled_data: dask array
        corrected dask array object reshaped on the detector plane, e.g. for a single frame case
        as above: (1, 512, 512)
    """
    width = hdr_info["width"]
    height = hdr_info["height"]
    width_height = width * height
    if (
        hdr_info["Counter Depth (number)"] == 24
        or hdr_info["Counter Depth (number)"] == 12
    ):
        cols = 4

    elif hdr_info["Counter Depth (number)"] == 1:
        cols = 64

    elif hdr_info["Counter Depth (number)"] == 6:
        cols = 8

    data = data.reshape((stack_size * width_height))

    data = data.reshape(stack_size, height * (height // cols), cols)

    data = da.flip(data, 2)

    if hdr_info["Assembly Size"] == "2x2":
        data = data.reshape((stack_size * width_height))
        data = data.reshape(stack_size, 512 // 2, 512 * 2)

        det1 = data[:, :, 0:256]
        det2 = data[:, :, 256:512]
        det3 = data[:, :, 512 : 512 + 256]
        det4 = data[:, :, 512 + 256 :]

        det3 = da.flip(det3, 2)
        det3 = da.flip(det3, 1)

        det4 = da.flip(det4, 2)
        det4 = da.flip(det4, 1)

        untangled_data = da.concatenate(
            (da.concatenate((det1, det3), 1), da.concatenate((det2, det4), 1)), 2
        )
    return untangled_data


def reshape_4DSTEM_FlyBack(data):
    """Reshapes the lazy-imported frame stack to navigation dimensions determined
    based on stored exposure times.


    Parameters
    ----------
    data : pyxem / hyperspy lazy Signal2D
        Lazy loaded electron diffraction data: <framenumbers | det_size, det_size>
        the data metadata contains flyback info as:
            ├── General
        │   └── title =
        └── Signal
            ├── binned = False
            ├── exposure_time = 0.001
            ├── flyback_times = [0.01826, 0.066, 0.065]
            ├── frames_number_skipped = 68
            ├── scan_X = 256
            └── signal_type = STEM

    Returns
    -------
    data_skip : pyxem.signals.LazyElectronDiffraction2D
        Reshaped electron diffraction data <scan_x, scan_y | det_size, det_size>
    """
    # Get detector size in pixels
    # detector size in pixels
    det_size = data.axes_manager[1].size
    # Read metadata
    skip_ind = data.metadata.Signal.frames_number_skipped
    line_len = data.metadata.Signal.scan_X

    n_lines = floor((data.data.shape[0] - skip_ind) / line_len)

    # Remove skipped frames
    data_skip = data.inav[skip_ind : skip_ind + (n_lines * line_len)]
    # Reshape signal
    data_skip.data = data_skip.data.reshape(n_lines, line_len, det_size, det_size)
    data_skip.axes_manager._axes.insert(0, data_skip.axes_manager[0].copy())
    data_skip.get_dimensions_from_data()
    # Cropping the bright fly-back pixel
    data_skip = data_skip.inav[1:]

    return data_skip


def reshape_4DSTEM_SumFrames(data):
    """
    Reshapes the lazy-imported stack of dimensions: (xxxxxx|Det_X, Det_Y) to the correct scan pattern
    shape: (x, y | Det_X, Det_Y) when the frame exposure times are not in headre bits.
    It utilises the over-exposed fly-back frame to identify the start of the lines in the first 20
    lines of frames,checks line length consistancy and finds the number of frames to skip at the
    beginning (this number is printed out as string output).

    Parameters
    ----------
    data : pyxem LazyElectronDiffraction2D
        imported mib file with diensions of: framenumbers|Det_X, Det_Y

    Returns
    -------
    data_reshaped : pyxem.signals.LazyElectronDiffraction2D
        reshaped data (x, y | Det_X, Det_Y)
    """
    # Assuming sacn_x, i.e. number of probe positions in a line is square root
    # of total number of frames
    scan_x = int(np.sqrt(data.axes_manager[0].size))
    # detector size in pixels
    det_size = data.axes_manager[1].size
    # crop the first ~20 lines
    data_crop = data.inav[0 : 20 * scan_x]
    data_crop_t = data_crop.T
    data_crop_t_sum = data_crop_t.sum()
    # summing over patterns
    intensity_array = data_crop_t_sum.data
    intensity_array = intensity_array.compute()
    peaks = find_peaks(intensity_array, distance=scan_x)
    # Difference between consecutive elements of the array
    lines = np.ediff1d(peaks[0])
    # Assuming the last element to be the line length
    line_len = lines[-1]
    if line_len != scan_x:
        raise ValueError("Fly_back does not correspond to correct line length!")
        #  Exits this routine and reshapes using scan size instead
    # Checking line lengths
    check = np.ravel(np.where(lines == line_len, True, False))
    # In case there is a False in there take the index of the last False
    if ~np.all(check):
        start_ind = np.where(check == False)[0][-1] + 2
        # Adding 2 - instead of 1 -  to be sure scan is OK
        skip_ind = peaks[0][start_ind]
    # In case they are all True take the index of the first True
    else:
        # number of frames to skip at the beginning
        skip_ind = peaks[0][0]
    # Number of lines
    n_lines = floor((data.data.shape[0] - skip_ind) / line_len)
    # with the skipped frames removed
    data_skip = data.inav[skip_ind : skip_ind + (n_lines * line_len)]
    data_skip.data = data_skip.data.reshape(n_lines, line_len, det_size, det_size)
    data_skip.axes_manager._axes.insert(0, data_skip.axes_manager[0].copy())
    data_skip.get_dimensions_from_data()
    print("Reshaping using the frame intensity sums of the first 20 lines")
    print("Number of frames skipped at the beginning: ", skip_ind)
    # Cropping the flyaback pixel at the start
    data_skip = data_skip.inav[1:]
    return data_skip



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from pyxem.signals.electron_diffraction1d import ElectronDiffraction1D
from pyxem.signals.reduced_intensity1d import ReducedIntensity1D
from pyxem.generators.red_intensity_generator1d import ReducedIntensityGenerator1D


@pytest.fixture
def red_int_generator():
    data = np.arange(10, 0, -1).reshape(1, 10) * np.arange(1, 5).reshape(4, 1)
    data = data.reshape(2, 2, 10)
    rp = ElectronDiffraction1D(data)
    rigen = ReducedIntensityGenerator1D(rp)
    return rigen


def test_init(red_int_generator):
    assert isinstance(red_int_generator, ReducedIntensityGenerator1D)


def test_scattering_calibration(red_int_generator):
    calib = 0.1
    red_int_generator.set_diffraction_calibration(calibration=calib)
    s_scale = red_int_generator.signal.axes_manager.signal_axes[0].scale
    assert s_scale == calib
    offset = 0.05
    red_int_generator.set_diffraction_offset(offset=offset)
    s_offset = red_int_generator.signal.axes_manager.signal_axes[0].offset
    assert s_offset == offset


def test_fit_atomic_scattering(red_int_generator):
    calib = 0.1
    red_int_generator.set_diffraction_calibration(calibration=calib)
    elements = ["Cu"]
    fracs = [1]
    assert red_int_generator.background_fit is None
    assert red_int_generator.normalisation is None
    red_int_generator.fit_atomic_scattering(elements=elements, fracs=fracs)

    fit_expected = np.array(
        [
            [
                [
                    10.3622224,
                    9.50839228,
                    7.75020528,
                    6.17950632,
                    5.09341816,
                    4.3789069,
                    3.89305285,
                    3.54593791,
                    3.28835785,
                    3.09296379,
                ],
                [
                    20.72444485,
                    19.01678459,
                    15.50041059,
                    12.35901266,
                    10.18683634,
                    8.75781381,
                    7.78610572,
                    7.09187582,
                    6.5767157,
                    6.18592759,
                ],
            ],
            [
                [
                    31.08666726,
                    28.52517688,
                    23.25061587,
                    18.53851898,
                    15.2802545,
                    13.1367207,
                    11.67915856,
                    10.63781372,
                    9.86507354,
                    9.27889137,
                ],
                [
                    41.44888967,
                    38.03356917,
                    31.00082115,
                    24.71802531,
                    20.37367266,
                    17.5156276,
                    15.57221142,
                    14.18375163,
                    13.15343139,
                    12.37185516,
                ],
            ],
        ]
    )
    norm_expected = np.array(
        [
            [
                [
                    7.96649232,
                    7.11266219,
                    5.35447519,
                    3.78377624,
                    2.69768808,
                    1.98317681,
                    1.49732277,
                    1.15020782,
                    0.89262776,
                    0.6972337,
                ],
                [
                    15.93298467,
                    14.22532442,
                    10.70895041,
                    7.56755249,
                    5.39537617,
                    3.96635363,
                    2.99464554,
                    2.30041564,
                    1.78525553,
                    1.39446741,
                ],
            ],
            [
                [
                    23.89947701,
                    21.33798663,
                    16.06342562,
                    11.35132873,
                    8.09306425,
                    5.94953045,
                    4.49196831,
                    3.45062347,
                    2.67788329,
                    2.09170112,
                ],
                [
                    31.86596934,
                    28.45064883,
                    21.41790082,
                    15.13510497,
                    10.79075233,
                    7.93270726,
                    5.98929108,
                    4.60083129,
                    3.57051106,
                    2.78893482,
                ],
            ],
        ]
    )

    assert np.allclose(red_int_generator.background_fit, fit_expected)
    assert np.allclose(red_int_generator.normalisation, norm_expected)


def test_set_cutoff(red_int_generator):
    s_min, s_max = 0, 8
    red_int_generator.set_s_cutoff(s_min, s_max)
    assert red_int_generator.cutoff == [s_min, s_max]


def test_subtract_bkgd(red_int_generator):
    bkgd_pattern = np.arange(10, 0, -1).reshape(10)
    red_int_generator.subtract_bkgd_pattern(bkgd_pattern)

    expected = np.arange(10, 0, -1).reshape(1, 10) * np.arange(0, 4).reshape(4, 1)
    expected = expected.reshape(2, 2, 10)

    assert np.array_equal(red_int_generator.signal.data, expected)


def test_mask_from_bkgd(red_int_generator):
    mask_pattern = np.arange(10, 0, -1).reshape(10)
    mask_threshold = 6.5
    red_int_generator.mask_from_bkgd_pattern(
        mask_pattern, mask_threshold=mask_threshold
    )

    expected = np.arange(10, 0, -1).reshape(1, 10) * np.arange(1, 5).reshape(4, 1)
    expected = expected.reshape(2, 2, 10)
    expected[:, :, :4] = 0

    assert np.array_equal(red_int_generator.signal.data, expected)


def test_mask_reduced_intensity(red_int_generator):
    mask_pattern = np.ones(10)
    mask_pattern[:4] = 0
    red_int_generator.mask_reduced_intensity(mask_pattern)

    expected = np.arange(10, 0, -1).reshape(1, 10) * np.arange(1, 5).reshape(4, 1)
    expected = expected.reshape(2, 2, 10)
    expected[:, :, :4] = 0

    assert np.array_equal(red_int_generator.signal.data, expected)


def test_incorrect_mask(red_int_generator):
    mask_pattern = np.ones(10)
    mask_pattern[:4] = 0
    mask_pattern[8] = 2
    with pytest.raises(
        ValueError, match="Masking array does not consist of zeroes and ones"
    ):
        red_int_generator.mask_reduced_intensity(mask_pattern)


def test_get_reduced_intensity(red_int_generator):
    red_int_generator.set_diffraction_calibration(calibration=0.1)
    red_int_generator.fit_atomic_scattering(elements=["Cu"], fracs=[1])
    ri = red_int_generator.get_reduced_intensity()
    assert isinstance(ri, ReducedIntensity1D)

    ri_expected_single = np.array(
        [
            -0.0,
            -0.04491037,
            0.0586241,
            0.40874355,
            0.8446079,
            0.98388682,
            0.26926809,
            -2.08758824,
            -7.25497587,
            -16.97484119,
        ]
    )
    ri_expected_array = np.vstack(
        (ri_expected_single, ri_expected_single, ri_expected_single, ri_expected_single)
    ).reshape(2, 2, 10)
    ri_expected = ReducedIntensity1D(ri_expected_array)

    assert np.allclose(ri.data, ri_expected.data)



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from pyxem.generators.subpixelrefinement_generator import (
    SubpixelrefinementGenerator,
    get_simulated_disc,
)
from pyxem.signals.diffraction_vectors import DiffractionVectors
from pyxem.signals.electron_diffraction2d import ElectronDiffraction2D
from skimage import draw


# @pytest.mark.xfail(raises=ValueError)
class Test_init_xfails:
    def test_out_of_range_vectors_numpy(self):
        """Test that putting vectors that lie outside of the
        diffraction patterns raises a ValueError"""
        vector = np.array([[1, -100]])
        dp = ElectronDiffraction2D(np.ones((20, 20)))

        with pytest.raises(
            ValueError,
            match="Some of your vectors do not lie within your diffraction pattern",
        ):
            sprg = SubpixelrefinementGenerator(dp, vector)

    def test_out_of_range_vectors_DiffractionVectors(self):
        """Test that putting vectors that lie outside of the
        diffraction patterns raises a ValueError"""
        vectors = DiffractionVectors(np.array([[1, -100]]))
        dp = ElectronDiffraction2D(np.ones((20, 20)))

        with pytest.raises(
            ValueError,
            match="Some of your vectors do not lie within your diffraction pattern",
        ):
            sprg = SubpixelrefinementGenerator(dp, vectors)

    def test_wrong_navigation_dimensions(self):
        """Tests that navigation dimensions must be appropriate too."""
        dp = ElectronDiffraction2D(np.zeros((2, 2, 8, 8)))
        vectors = DiffractionVectors(np.zeros((1, 2)))
        dp.axes_manager.set_signal_dimension(2)
        vectors.axes_manager.set_signal_dimension(0)

        # Note - uses regex via re.search()
        with pytest.raises(
            ValueError,
            match=r"Vectors with shape .* must have the same navigation shape as .*",
        ):
            sprg = SubpixelrefinementGenerator(dp, vectors)


class set_up_for_subpixelpeakfinders:
    def create_spot(self):
        z1, z1a = np.zeros((128, 128)), np.zeros((128, 128))
        z2, z2a = np.zeros((128, 128)), np.zeros((128, 128))

        rr, cc = draw.circle(30, 90, radius=4, shape=z1.shape)  # 30 is y!
        z1[rr, cc], z2[rr, cc] = 1, 1
        rr2, cc2 = draw.circle(100, 60, radius=4, shape=z2.shape)
        z2[rr2, cc2] = 1
        rr, cc = draw.circle(30, 90 + 3, radius=4, shape=z1.shape)  # 30 is y!
        z1a[rr, cc], z2a[rr, cc] = 1, 1
        rr2, cc2 = draw.circle(100 - 2, 60, radius=4, shape=z2.shape)
        z2a[rr2, cc2] = 1

        # marks centers for local com and local_gaussian_method
        z1[30, 90], z2[30, 90], z2[100, 60] = 2, 2, 2
        z1a[30, 93], z2a[30, 93], z2a[98, 60] = 10, 10, 10

        dp = ElectronDiffraction2D(
            np.asarray([[z1, z1a], [z2, z2a]])
        )  # this needs to be in 2x2
        return dp

    def create_Diffraction_vectors(self):
        v1 = np.array([[90 - 64, 30 - 64]])
        v2 = np.array([[90 - 64, 30 - 64], [100 - 64, 60 - 64]])
        vectors = DiffractionVectors(np.array([[v1, v1], [v2, v2]]))
        vectors.axes_manager.set_signal_dimension(0)
        return vectors


class Test_subpixelpeakfinders:
    """ Tests the various peak finders have the correct x,y conventions for
    both the vectors and the shifts, in both the numpy and the DiffractionVectors
    cases as well as confirming we have avoided 'off by one' errors """

    set_up = set_up_for_subpixelpeakfinders()

    @pytest.fixture(
        params=[set_up.create_Diffraction_vectors(), np.array([[90 - 64, 30 - 64]])]
    )
    def diffraction_vectors(self, request):
        # see https://bit.ly/2mXpSlD for an example of this architecture
        return request.param

    def get_spr(self, diffraction_vectors):
        dp = set_up_for_subpixelpeakfinders().create_spot()
        return SubpixelrefinementGenerator(dp, diffraction_vectors)

    def no_shift_case(self, s):
        error = s.data[0, 0] - np.asarray([[90 - 64, 30 - 64]])
        rms_error = np.sqrt(error[0, 0] ** 2 + error[0, 1] ** 2)
        assert rms_error < 1e-5  # perfect detection for this trivial case

    def x_shift_case(self, s):
        error = s.data[0, 1] - np.asarray([[93 - 64, 30 - 64]])
        rms_error = np.sqrt(error[0, 0] ** 2 + error[0, 1] ** 2)
        assert rms_error < 0.5  # correct to within a pixel

    def test_assertioned_xc(self, diffraction_vectors):
        subpixelsfound = self.get_spr(diffraction_vectors).conventional_xc(12, 4, 8)
        self.no_shift_case(subpixelsfound)
        self.x_shift_case(subpixelsfound)

    def test_assertioned_com(self, diffraction_vectors):
        subpixelsfound = self.get_spr(diffraction_vectors).center_of_mass_method(12)
        self.no_shift_case(subpixelsfound)
        self.x_shift_case(subpixelsfound)

    def test_assertioned_log(self, diffraction_vectors):
        with pytest.warns(
            UserWarning,
            match="peak in your pattern that lies on the edge of the square",
        ):
            subpixelsfound = self.get_spr(diffraction_vectors).local_gaussian_method(12)

        self.no_shift_case(subpixelsfound)
        self.x_shift_case(subpixelsfound)


# class Test_misc():
""" These tests will be removed for 0.11.0, but are needed for the log method &
security on the x/y conventions until then """


def create_spot_gaussian():
    z1 = np.zeros((128, 128))
    x = np.arange(0.0, 10, 1.0)
    y = x[:, np.newaxis]
    z1[20:30, 50:60] = np.exp(-((x - 5.1) ** 2 + (y - 5.3) ** 2) / 4)
    dp = ElectronDiffraction2D(
        np.asarray([[z1, z1], [z1, z1]])
    )  # this needs to be in 2x2
    return dp


@pytest.mark.parametrize(
    "dp, diffraction_vectors",
    [(create_spot_gaussian(), np.array([[55 - 64, 25 - 64]]))],
)
@pytest.mark.filterwarnings("ignore::UserWarning")  # our warning
def test_bad_square_size_local_gaussian_method(dp, diffraction_vectors):
    spr = SubpixelrefinementGenerator(dp, diffraction_vectors)
    s = spr.local_gaussian_method(2)


def test_xy_errors_in_conventional_xc_method_as_per_issue_490():
    """ This was the MWE example code for the issue """
    dp = get_simulated_disc(100, 20)
    # translate y by +4
    shifted = np.pad(dp, ((0, 4), (0, 0)), "constant")[4:].reshape(1, 1, *dp.shape)
    signal = ElectronDiffraction2D(shifted)
    spg = SubpixelrefinementGenerator(signal, np.array([[0, 0]]))
    peaks = spg.conventional_xc(100, 20, 1).data[0, 0, 0]  # as quoted in the issue
    np.testing.assert_allclose([0, -4], peaks)
    """ we also test com method for clarity """
    peaks = spg.center_of_mass_method(60).data[0, 0, 0]
    np.testing.assert_allclose([0, -4], peaks, atol=1.5)
    """ we also test reference_xc """
    peaks = spg.reference_xc(100, dp, 1).data[0, 0, 0]  # as quoted in the issue
    np.testing.assert_allclose([0, -4], peaks)



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np
import pyxem as pxm
import os

from pyxem.signals.crystallographic_map import CrystallographicMap
from pyxem.signals.electron_diffraction1d import ElectronDiffraction1D
from pyxem.signals.electron_diffraction2d import ElectronDiffraction2D
from pyxem.signals.diffraction_vectors import DiffractionVectors, DiffractionVectors2D
from pyxem.signals.indexation_results import TemplateMatchingResults
from pyxem.signals.vdf_image import VDFImage


@pytest.mark.parametrize(
    "class_to_test,meta_string",
    [
        (ElectronDiffraction2D, "string1"),
        (TemplateMatchingResults, "string2"),
        (DiffractionVectors, "string3"),
        (DiffractionVectors2D, "string3"),
        (CrystallographicMap, "string4"),
        (ElectronDiffraction1D, "string5"),
        (VDFImage, "string6"),
    ],
)
def test_load_function_core(class_to_test, meta_string):
    """
    Test the core; which is load a previously saved pyxem object.
    """
    to_save = class_to_test(np.zeros((2, 2, 2, 2)))
    to_save.metadata.Signal.tracker = meta_string
    if class_to_test is DiffractionVectors:
        to_save.axes_manager.set_signal_dimension(0)
    if class_to_test is DiffractionVectors2D:
        to_save.axes_manager.set_signal_dimension(2)
    to_save.save("tempfile_for_load_and_save.hspy", overwrite=True)
    from_save = pxm.load("tempfile_for_load_and_save.hspy")
    assert isinstance(from_save, class_to_test)
    assert from_save.metadata.Signal.tracker == meta_string
    assert np.allclose(to_save.data, from_save.data)
    os.remove("tempfile_for_load_and_save.hspy")


@pytest.fixture()
def make_saved_dp(diffraction_pattern):
    """
    This makes use of conftest
    """
    diffraction_pattern.save("dp_temp")
    yield
    os.remove("dp_temp.hspy")


def test_load_ElectronDiffraction2D(diffraction_pattern, make_saved_dp):
    """
    This tests that our load function keeps .data, instance and metadata
    """
    dp = pxm.load("dp_temp.hspy")
    assert np.allclose(dp.data, diffraction_pattern.data)
    assert isinstance(dp, ElectronDiffraction2D)
    assert (
        diffraction_pattern.metadata.Signal.found_from == dp.metadata.Signal.found_from
    )



# -*- coding: utf-8 -*-
# Copyright 2019 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np

from pyxem.utils.indexation_utils import (
    crystal_from_template_matching,
    crystal_from_vector_matching,
    match_vectors,
    zero_mean_normalized_correlation,
    fast_correlation,
)


def test_zero_mean_normalized_correlation():
    np.testing.assert_approx_equal(
        zero_mean_normalized_correlation(
            3, np.linalg.norm([2 / 3, 1 / 3, 1 / 3]), 1 / 3, [1, 0, 0], [1, 0, 0]
        ),
        1,
    )
    # nb_pixels,image_std,average_image_intensity,image_intensities,int_local
    assert zero_mean_normalized_correlation(3, 0, 1, [1, 1, 1], [0, 0, 1]) == 0


def test_fast_correlation():
    np.testing.assert_approx_equal(
        fast_correlation([1, 1, 1], [1, 1, 1], np.sqrt(3)), np.sqrt(3)
    )
    np.testing.assert_approx_equal(fast_correlation([1, 1, 1], [1, 0, 0], 1), 1)


def test_crystal_from_template_matching_sp(sp_template_match_result):
    # branch single phase
    cmap = crystal_from_template_matching(sp_template_match_result)
    assert cmap[0] == 0
    np.testing.assert_allclose(cmap[1], [2, 3, 4])
    np.testing.assert_allclose(cmap[2]["correlation"], 0.7)
    np.testing.assert_allclose(
        cmap[2]["orientation_reliability"], 100 * (1 - (0.6 / 0.7))
    )


def test_crystal_from_template_matching_dp(dp_template_match_result):
    # branch double phase
    cmap = crystal_from_template_matching(dp_template_match_result)
    r_or = 100 * (1 - (0.7 / 0.8))
    r_ph = 100 * (1 - (0.5 / 0.8))
    assert cmap[0] == 0
    np.testing.assert_allclose(cmap[1], [2, 3, 5])
    np.testing.assert_allclose(cmap[2]["correlation"], 0.8)
    np.testing.assert_allclose(cmap[2]["orientation_reliability"], r_or)
    np.testing.assert_allclose(cmap[2]["phase_reliability"], r_ph)


def test_crystal_from_vector_matching_sp(sp_vector_match_result):
    # branch single phase
    cmap = crystal_from_vector_matching(sp_vector_match_result)
    assert cmap[0] == 0
    np.testing.assert_allclose(cmap[1], [0, 0, 90])
    np.testing.assert_allclose(cmap[2]["match_rate"], 0.5)
    np.testing.assert_allclose(cmap[2]["ehkls"], np.array([0.1, 0.05, 0.2]))
    np.testing.assert_allclose(cmap[2]["total_error"], 0.1)
    np.testing.assert_allclose(
        cmap[2]["orientation_reliability"], 100 * (1 - (0.1 / 0.2))
    )


def test_crystal_from_vector_matching_dp(dp_vector_match_result):
    # branch double phase
    cmap = crystal_from_vector_matching(dp_vector_match_result)
    r_or = 100 * (1 - (0.1 / 0.2))
    r_ph = 100 * (1 - (0.1 / 0.3))
    assert cmap[0] == 1
    np.testing.assert_allclose(cmap[1], [0, 45, 45])
    np.testing.assert_allclose(cmap[2]["match_rate"], 0.8)
    np.testing.assert_allclose(cmap[2]["ehkls"], np.array([0.1, 0.3, 0.2]))
    np.testing.assert_allclose(cmap[2]["total_error"], 0.1)
    np.testing.assert_allclose(cmap[2]["orientation_reliability"], r_or)
    np.testing.assert_allclose(cmap[2]["phase_reliability"], r_ph)


def test_match_vectors(vector_match_peaks, vector_library):
    # Wrap to test handling of ragged arrays
    peaks = np.empty(1, dtype="object")
    peaks[0] = vector_match_peaks
    matches, rhkls = match_vectors(
        peaks,
        vector_library,
        mag_tol=0.1,
        angle_tol=0.1,
        index_error_tol=0.3,
        n_peaks_to_index=2,
        n_best=1,
    )
    assert len(matches) == 1
    np.testing.assert_allclose(matches[0].match_rate, 1.0)
    np.testing.assert_allclose(matches[0].rotation_matrix, np.identity(3), atol=0.1)
    np.testing.assert_allclose(matches[0].total_error, 0.03, atol=0.01)

    np.testing.assert_allclose(rhkls[0][0], [1, 0, 0])
    np.testing.assert_allclose(rhkls[0][1], [0, 2, 0])
    np.testing.assert_allclose(rhkls[0][2], [1, 2, 3])


def test_match_vector_total_error_default(vector_match_peaks, vector_library):
    matches, rhkls = match_vectors(
        vector_match_peaks,
        vector_library,
        mag_tol=0.1,
        angle_tol=0.1,
        index_error_tol=0.0,
        n_peaks_to_index=2,
        n_best=5,
    )
    assert len(matches) == 5
    np.testing.assert_allclose(matches[0][2], 0.0)  # match rate
    np.testing.assert_allclose(matches[0][1], np.identity(3), atol=0.1)
    np.testing.assert_allclose(matches[0][4], 1.0)  # error mean

    assert len(rhkls) == 0



# Copyright 2016-2020 The pyXem developers
# -*- coding: utf-8 -*-
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

"""
utils to support SubpixelrefinementGenerator
"""

import numpy as np
from skimage import draw
from skimage.transform import rescale
from pyxem.signals.diffraction_vectors import DiffractionVectors


def get_experimental_square(z, vector, square_size):
    """Defines a square region around a given diffraction vector and returns.

    Parameters
    ----------
    z : np.array()
        Single diffraction pattern
    vector : np.array()
        Single vector in pixels (int) [x,y] with top left as [0,0]
    square_size : int
        The length of one side of the bounding square (must be even)

    Returns
    -------
    square : np.array()
        Of size (L,L) where L = square_size

    """
    if square_size % 2 != 0:
        raise ValueError("'square_size' must be an even number")

    cx, cy, half_ss = vector[0], vector[1], int(square_size / 2)
    # select square with correct x,y see PR for details
    _z = z[cy - half_ss : cy + half_ss, cx - half_ss : cx + half_ss]
    return _z


def get_simulated_disc(square_size, disc_radius):
    """Create a uniform disc for correlating with the experimental square.

    Parameters
    ----------
    square size : int
        (even) - size of the bounding box
    disc_radius : int
        radius of the disc

    Returns
    -------
    arr: np.array()
        Upsampled copy of the simulated disc as a numpy array

    """

    if square_size % 2 != 0:
        raise ValueError("'square_size' must be an even number")

    ss = int(square_size)
    arr = np.zeros((ss, ss))
    rr, cc = draw.circle(
        int(ss / 2), int(ss / 2), radius=disc_radius, shape=arr.shape
    )  # is the thin disc a good idea
    arr[rr, cc] = 1
    return arr


def _get_pixel_vectors(dp, vectors, calibration, center):
    """Get the pixel coordinates for the given diffraction
    pattern and vectors.

    Parameters
    ----------
    dp: :obj:`pyxem.signals.ElectronDiffraction2D`
        Instance of ElectronDiffraction2D
    vectors : :obj:`pyxem.signals.diffraction_vectors.DiffractionVectors`
        List of diffraction vectors
    calibration : [float, float]
        Calibration values
    center : float, float
        Image origin in pixel coordinates

    Returns
    -------
    vector_pixels : :obj:`pyxem.signals.diffraction_vectors.DiffractionVectors`
        Pixel coordinates for given diffraction pattern and vectors.
    """

    def _floor(vectors, calibration, center):
        if vectors.shape == (1,) and vectors.dtype == np.object:
            vectors = vectors[0]
        return np.floor((vectors.astype(np.float64) / calibration) + center).astype(
            np.int
        )

    if isinstance(vectors, DiffractionVectors):
        if vectors.axes_manager.navigation_shape != dp.axes_manager.navigation_shape:
            raise ValueError(
                "Vectors with shape {} must have the same navigation shape "
                "as the diffraction patterns which has shape {}.".format(
                    vectors.axes_manager.navigation_shape,
                    dp.axes_manager.navigation_shape,
                )
            )
        vector_pixels = vectors.map(
            _floor, calibration=calibration, center=center, inplace=False
        )
    else:
        vector_pixels = _floor(vectors, calibration, center)

    if isinstance(vector_pixels, DiffractionVectors):
        if np.any(vector_pixels.data > (np.max(dp.data.shape) - 1)) or (
            np.any(vector_pixels.data < 0)
        ):
            raise ValueError(
                "Some of your vectors do not lie within your diffraction pattern, check your calibration"
            )
    elif isinstance(vector_pixels, np.ndarray):
        if np.any((vector_pixels > np.max(dp.data.shape) - 1)) or (
            np.any(vector_pixels < 0)
        ):
            raise ValueError(
                "Some of your vectors do not lie within your diffraction pattern, check your calibration"
            )

    return vector_pixels



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pytest
from pyxem.signals.crystallographic_map import CrystallographicMap
from pyxem.signals.crystallographic_map import load_mtex_map
from pyxem.signals.crystallographic_map import _distance_from_fixed_angle
from transforms3d.euler import euler2quat, quat2axangle
from transforms3d.quaternions import qmult, qinverse
import os


def worker_for_test_CrystallographicMap_io(mapp):
    mapp.save_mtex_map("file_01.txt")
    lmap = load_mtex_map("file_01.txt")
    os.remove("file_01.txt")
    return mapp, lmap


def get_distance_between_two_angles_longform(angle_1, angle_2):
    """
    Using the long form to find the distance between two angles in euler form
    """
    q1 = euler2quat(*np.deg2rad(angle_1), axes="rzxz")
    q2 = euler2quat(*np.deg2rad(angle_2), axes="rzxz")
    # now assume transform of the form MODAL then Something = TOTAL
    # so we want to calculate MODAL^{-1} TOTAL

    q_from_mode = qmult(qinverse(q2), q1)
    axis, angle = quat2axangle(q_from_mode)
    return np.rad2deg(angle)


@pytest.fixture()
def sp_cryst_map():
    """
    Generates a single phase Crystallographic Map
    """
    base = np.empty((4, 3), dtype="object")
    base[0] = [
        0,
        np.array([5, 17, 6]),
        {"correlation": 3e-17, "orientation_reliability": 0.5},
    ]
    base[1] = [
        0,
        np.array([6, 17, 6]),
        {"correlation": 2e-17, "orientation_reliability": 0.4},
    ]
    base[2] = [
        0,
        np.array([12, 3, 6]),
        {"correlation": 4e-17, "orientation_reliability": 0.3},
    ]
    base[3] = [
        0,
        np.array([12, 3, 5]),
        {"correlation": 8e-16, "orientation_reliability": 0.2},
    ]
    crystal_map = CrystallographicMap(base.reshape((2, 2, 3)))
    crystal_map.method = "template_matching"
    return crystal_map


@pytest.fixture()
def dp_cryst_map():
    """
    Generates a Crystallographic Map with two phases
    """
    base = np.empty((4, 3), dtype="object")
    base[0] = [
        0,
        np.array([5, 17, 6]),
        {
            "correlation": 3e-17,
            "orientation_reliability": 0.5,
            "phase_reliability": 0.6,
        },
    ]
    base[1] = [
        1,
        np.array([6, 17, 6]),
        {
            "correlation": 2e-17,
            "orientation_reliability": 0.4,
            "phase_reliability": 0.7,
        },
    ]
    base[2] = [
        0,
        np.array([12, 3, 6]),
        {
            "correlation": 4e-17,
            "orientation_reliability": 0.3,
            "phase_reliability": 0.1,
        },
    ]
    base[3] = [
        0,
        np.array([12, 3, 5]),
        {
            "correlation": 8e-16,
            "orientation_reliability": 0.2,
            "phase_reliability": 0.8,
        },
    ]
    crystal_map = CrystallographicMap(base.reshape((2, 2, 3)))
    crystal_map.method = "template_matching"
    return crystal_map


@pytest.fixture()
def mod_cryst_map():
    """
    Generates a Crystallographic Map with (5,17,6) as the modal angle
    """
    base = np.empty((6, 3), dtype="object")
    base[0] = [
        0,
        np.array([5, 17, 6]),
        {"correlation": 5e-17, "orientation_reliability": 0.5},
    ]
    base[1] = [
        0,
        np.array([5, 17, 6]),
        {"correlation": 5e-17, "orientation_reliability": 0.5},
    ]
    base[2] = [
        0,
        np.array([6, 19, 6]),
        {"correlation": 5e-17, "orientation_reliability": 0.5},
    ]
    base[3] = [
        0,
        np.array([7, 19, 6]),
        {"correlation": 5e-17, "orientation_reliability": 0.5},
    ]
    base[4] = [
        0,
        np.array([8, 19, 6]),
        {"correlation": 5e-17, "orientation_reliability": 0.5},
    ]
    base[5] = [
        0,
        np.array([9, 19, 6]),
        {"correlation": 5e-17, "orientation_reliability": 0.5},
    ]
    crystal_map = CrystallographicMap(base.reshape((3, 2, 3)))
    crystal_map.method = "template_matching"
    return crystal_map


@pytest.fixture()
def dp_cryst_map_vector():
    """
    Generates a Crystallographic Map with two phases from vector matching
    """
    base = np.empty((4, 3), dtype="object")
    base[0] = [
        0,
        np.array([5, 17, 6]),
        {
            "match_rate": 0.5,
            "ehkls": np.array([0.1, 0.05, 0.2]),
            "total_error": 0.1,
            "orientation_reliability": 13.2,
            "phase_reliability": 42.0,
        },
    ]
    base[1] = [
        1,
        np.array([6, 17, 6]),
        {
            "match_rate": 0.5,
            "ehkls": np.array([0.1, 0.05, 0.2]),
            "total_error": 0.1,
            "orientation_reliability": 13.2,
            "phase_reliability": 42.0,
        },
    ]
    base[2] = [
        0,
        np.array([12, 3, 6]),
        {
            "match_rate": 0.5,
            "ehkls": np.array([0.1, 0.05, 0.2]),
            "total_error": 0.1,
            "orientation_reliability": 13.2,
            "phase_reliability": 42.0,
        },
    ]
    base[3] = [
        0,
        np.array([12, 3, 5]),
        {
            "match_rate": 0.5,
            "ehkls": np.array([0.1, 0.05, 0.2]),
            "total_error": 0.1,
            "orientation_reliability": 13.2,
            "phase_reliability": 42.0,
        },
    ]
    crystal_map = CrystallographicMap(base.reshape((2, 2, 3)))
    crystal_map.method = "vector_matching"
    return crystal_map


class TestMapCreation:
    def test_get_phase_map(self, sp_cryst_map):
        phasemap = sp_cryst_map.get_phase_map()
        assert phasemap.isig[0, 0] == 0

    def test_get_orientation_map(self, sp_cryst_map):
        orimap = sp_cryst_map.get_orientation_map()
        assert orimap.isig[0, 0] == 0

    @pytest.mark.parametrize(
        "metric, value",
        [
            ("correlation", 3e-17),
            ("orientation_reliability", 0.5),
            ("phase_reliability", 0.6),
        ],
    )
    def test_get_metric_map_template_match(self, dp_cryst_map, metric, value):
        metric_map = dp_cryst_map.get_metric_map(metric)
        assert metric_map.isig[0, 0] == value

    @pytest.mark.parametrize(
        "metric, value",
        [
            ("match_rate", 0.5),
            ("ehkls", np.array([0.1, 0.05, 0.2])),
            ("total_error", 0.1),
            ("orientation_reliability", 13.2),
            ("phase_reliability", 42.0),
        ],
    )
    def test_get_metric_map_vector_match(self, dp_cryst_map_vector, metric, value):
        metric_map = dp_cryst_map_vector.get_metric_map(metric)
        assert np.allclose(metric_map.isig[0, 0], value)

    def test_get_metric_map_template_match_bad_metric(self, sp_cryst_map):
        # Note - uses regex via re.search()
        with pytest.raises(
            ValueError, match=r"metric .* is not valid for template matching"
        ):
            metric_map = sp_cryst_map.get_metric_map("no metric")

    def test_get_metric_map_vector_match_bad_metric(self, dp_cryst_map_vector):
        # Note - uses regex via re.search()
        with pytest.raises(
            ValueError, match=r"metric .* is not valid for vector matching"
        ):
            metric_map = dp_cryst_map_vector.get_metric_map("no metric")

    def test_get_metric_map_no_method(self):
        crystal_map = CrystallographicMap(np.array([[1]]))

        with pytest.raises(
            ValueError, match="crystallographic mapping method must be specified"
        ):
            metric_map = crystal_map.get_metric_map("no metric")


class TestMTEXIO:
    def test_Crystallographic_Map_io_sp(self, sp_cryst_map):
        saved, loaded = worker_for_test_CrystallographicMap_io(sp_cryst_map)
        saved.method = "template_matching"
        loaded.method = "template_matching"
        np.testing.assert_allclose(
            saved.isig[0].data.astype(np.int), loaded.isig[0].data.astype(np.int)
        )
        np.testing.assert_allclose(
            np.array(saved.isig[1].data.tolist()),
            np.array(loaded.isig[1].data.tolist()),
        )
        np.testing.assert_allclose(
            saved.get_metric_map("correlation").data,
            loaded.get_metric_map("correlation").data,
        )

    def test_Crystallographic_Map_io_dp(self, dp_cryst_map):
        saved, loaded = worker_for_test_CrystallographicMap_io(dp_cryst_map)
        saved.method = "template_matching"
        loaded.method = "template_matching"
        np.testing.assert_allclose(
            saved.isig[0].data.astype(np.int), loaded.isig[0].data.astype(np.int)
        )
        np.testing.assert_allclose(
            np.array(saved.isig[1].data.tolist()),
            np.array(loaded.isig[1].data.tolist()),
        )
        np.testing.assert_allclose(
            saved.get_metric_map("correlation").data,
            loaded.get_metric_map("correlation").data,
        )


class TestModalAngularFunctionality:
    def test_get_distance_from_modal(self, mod_cryst_map):
        # function runs without error
        formal = mod_cryst_map.get_distance_from_modal_angle()
        assert True

    def test_get_modal_angles(self, mod_cryst_map):
        # modal angle is found correctly
        out = mod_cryst_map.get_modal_angles()
        assert np.allclose(out[0], [5, 17, 6])
        assert np.allclose(out[1], (2 / 6))

    def test_get_distance_from_fixed_angle(self):
        # distance between two angles is found correctly
        angle_1 = [1, 1, 3]
        angle_2 = [1, 1, 4]
        implemented = _distance_from_fixed_angle([angle_1], angle_2)
        testing = get_distance_between_two_angles_longform(angle_1, angle_2)
        assert np.allclose(implemented, testing)
        assert np.allclose(implemented, 1)



# -*- coding: utf-8 -*-
# Copyright 2016-2020 The pyXem developers
#
# This file is part of pyXem.
#
# pyXem is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pyXem is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyXem.  If not, see <http://www.gnu.org/licenses/>.

import pytest
import numpy as np

from pyxem.generators.vdf_generator import VDFGenerator

from pyxem.signals.electron_diffraction2d import ElectronDiffraction2D
from pyxem.signals.diffraction_vectors import DiffractionVectors
from pyxem.signals.vdf_image import VDFImage


@pytest.fixture(params=[np.array([[1, 1], [2, 2]])])
def diffraction_vectors(request):
    dvec = DiffractionVectors(request.param)
    dvec.axes_manager.set_signal_dimension(1)
    return dvec


@pytest.fixture
def vdf_generator(diffraction_pattern, diffraction_vectors):
    diffraction_pattern.data = np.where(
        diffraction_pattern.data == 0, 0.01, diffraction_pattern.data
    )  # avoid divide by zeroes
    return VDFGenerator(diffraction_pattern, diffraction_vectors)


class TestVDFGenerator:
    def test_vdf_generator_init_with_vectors(self, diffraction_pattern):
        dvm = DiffractionVectors(
            np.array(
                [
                    [np.array([[1, 1], [2, 2]]), np.array([[1, 1], [2, 2], [1, 2]])],
                    [np.array([[1, 1], [2, 2]]), np.array([[1, 1], [2, 2]])],
                ],
                dtype=object,
            )
        )
        dvm.axes_manager.set_signal_dimension(0)

        vdfgen = VDFGenerator(diffraction_pattern, dvm)
        assert isinstance(vdfgen.signal, ElectronDiffraction2D)
        assert isinstance(vdfgen.vectors, DiffractionVectors)

    def test_vdf_generator_init_without_vectors(self, diffraction_pattern):

        vdfgen = VDFGenerator(diffraction_pattern)
        assert isinstance(vdfgen.signal, ElectronDiffraction2D)
        assert isinstance(vdfgen.vectors, type(None))

    def test_vector_vdfs_without_vectors(self, diffraction_pattern):
        vdfgen = VDFGenerator(diffraction_pattern)
        with pytest.raises(
            ValueError, match="DiffractionVectors not specified by user"
        ):
            vdfgen.get_vector_vdf_images(radius=2.0)

    @pytest.mark.parametrize("radius, normalize", [(4.0, False), (4.0, True)])
    def test_get_vector_vdf_images(
        self, vdf_generator: VDFGenerator, radius, normalize
    ):
        vdfs = vdf_generator.get_vector_vdf_images(radius, normalize)
        assert isinstance(vdfs, VDFImage)

    @pytest.mark.parametrize(
        "k_min, k_max, k_steps, normalize", [(0.0, 4.0, 2, False), (0.0, 4.0, 2, True)]
    )
    def test_get_concentric_vdf_images(
        self, vdf_generator: VDFGenerator, k_min, k_max, k_steps, normalize
    ):
        vdfs = vdf_generator.get_concentric_vdf_images(k_min, k_max, k_steps, normalize)
        assert isinstance(vdfs, VDFImage)


def test_vdf_generator_from_map(diffraction_pattern):
    dvm = DiffractionVectors(
        np.array(
            [
                [np.array([[1, 1], [2, 2]]), np.array([[1, 1], [2, 2], [1, 2]])],
                [np.array([[1, 1], [2, 2]]), np.array([[1, 1], [2, 2]])],
            ],
            dtype=object,
        )
    )
    dvm.axes_manager.set_signal_dimension(0)

    vdfgen = VDFGenerator(diffraction_pattern, dvm)
    assert isinstance(vdfgen, VDFGenerator)
